# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
A module that assists with implementing the Discord OAuth2 flow.

.. currentmodule:: curious.oauth
"""
import collections
import datetime
import enum
import secrets
import typing

from asks.sessions import Session
from oauthlib.oauth2 import OAuth2Error
from oauthlib.oauth2.rfc6749.clients import WebApplicationClient


class InvalidStateError(Exception):
    """
    Raised from :meth:`.OAuth2Handshaker.fetch_token` if the state is invalid
    """
    pass


class OAuth2Scope(enum.Enum):
    """
    OAuth2 scopes.
    """
    #: Authorizes a bot into a guild.
    BOT = 'bot'
    #: Allows access to the connections of a user.
    CONNECTIONS = 'connections'
    #: Allows access to basic user info.
    IDENTIFY = 'identify'
    #: Allows access to basic user info + their email.
    EMAIL = 'email'
    #: Allows access to user guild objects for this user.
    GUILDS = 'guilds'
    #: Allows forcibly joining a guild on the behalf of a user.
    GUILDS_JOIN = 'guilds.join'
    #: Allows forcibly adding users to group DMs.
    GDM_JOIN = 'gdm.join'

    #: Allows reading messages via RPC.
    MESSAGES_READ = 'messages.read'
    #: Allows RPC client control.
    RPC = 'rpc'
    #: Allows RPC API control.
    RPC_API = 'rpc.api'
    #: Allows RPC notification reading.
    RPC_NOTIFICATIONS_READ = 'rpc.notifications.read'

    #: Creates an incoming webhook when authorized.
    WEBHOOK_INCOMING = 'webhook.incoming'


class OAuth2Token(object):
    """
    Represents a token returned from the Discord OAuth2 API.
    """

    def __init__(self, token_type: str, scope: str,
                 access_token: str, refresh_token: str, expiration_time: datetime.datetime):
        #: The token type of the token (normally ``Bearer``).
        self.token_type = token_type

        #: A list of :class:`.OAuth2Scope` this token is authenticated for.
        self.scopes = []
        for scope_name in scope.split(" "):
            self.scopes.append(OAuth2Scope(scope_name))

        #: The actual access token to be used.
        self.access_token = access_token

        #: The refresh token to be used during a refresh.
        self.refresh_token = refresh_token

        #: The time this token expires at.
        self.expiration_time = expiration_time

    @classmethod
    def from_dict(cls, d: dict) -> 'OAuth2Token':
        """
        Creates a token from a dict, similar to one provided by the token endpoint.
        """
        if "expiration_time" not in d:
            expiration_time = datetime.datetime.utcnow() + \
                              datetime.timedelta(seconds=d["expires_in"])
            d["expiration_time"] = expiration_time

        d.pop("expires_in", None)

        c = cls(**d)
        return c

    @property
    def expired(self) -> bool:
        return self.expiration_time < datetime.datetime.utcnow()

    def __repr__(self) -> str:
        return "<OAuth2Token access='{}' refresh='{}'>".format(self.access_token,
                                                               self.refresh_token)


class OAuth2Client(object):
    """
    The class used to perform an OAuth2 handshake with Discord.
    This will provide a URL that can be used to authorize a client, and then the ability to fetch a
    new :class:`.OAuth2Token`.
    
    .. code-block:: python3
    
        # note: we can't fetch the client ID automatically, unlike in regular bots :(
        # so you have to pass it manually
        # if you have a bot running alongside this, you can use the Client ID from that, however.
        my_client = OAuth2Client(my_client_id, my_client_secret)
        
        url = my_client.get_authorization_url(scopes=[OAuthScope.IDENTITY])
        # get the user to give back the state and the code
        # ...
        token = await my_client.get_token(state, code)
        
        # some time later
        user = await my_client.get_user()

    
    :param client_id:
        The Client ID of the application.
        
        .. warning::
            For **old bots**, this is NOT the Bot ID.
     
    :param client_secret: 
        The client secret of the application.
        
        ..warning::
            This is **not** your token.
            
    :param redirect_uri: 
        The URL the client will be redirected to after authorizing your application.  
        This is then used to retrieve the state.
    """

    BASE = "https://discordapp.com"
    API_BASE = "/api/v7"
    AUTHORIZE_URL = "/oauth2/authorize"
    TOKEN_URL = "/oauth2/token"

    def __init__(self, client_id: int, client_secret: str,
                 redirect_uri: str):
        self.client_id = client_id
        self.client_secret = client_secret
        self.redirect_uri = redirect_uri

        self._oauth2_client = WebApplicationClient(client_id=self.client_id,
                                                   redirect_url=self.redirect_uri)
        self.sess = Session(base_location=self.BASE, endpoint=self.API_BASE)

        #: A list of states that have been seen before.
        #: If the state was not seen, it will raise an invalid state error.
        self._states = collections.deque(maxlen=500)

    def _get_state(self):
        return secrets.token_urlsafe(16)

    def get_authorization_url(self, scopes: typing.List[OAuth2Scope]):
        """
        Gets the authorization URL to be used for the user to authorize this application.
        
        :param scopes: A list of :class:`.OAuthScope` for the request.
        """
        url = self._oauth2_client.prepare_request_uri(self.AUTHORIZE_URL,
                                                      scope=[scope.value for scope in scopes],
                                                      redirect_uri=self.redirect_uri,
                                                      state=self._get_state())
        return url

    async def fetch_token(self, code: str, state: str) -> OAuth2Token:
        """
        Fetches the token when given an authorization code and state.
        
        :param code: The authorization code returned in the URI. 
        :param state: The state returned in the URI.
        :return: A :class:`.OAuthToken` object representing the token.
        """
        # if state not in self._states:
        #    raise InvalidStateError(state)
        # construct the URI we need to send
        uri = self._oauth2_client.prepare_request_uri(self.TOKEN_URL,
                                                      code=code,
                                                      redirect_uri=self.redirect_uri,
                                                      client_secret=self.client_secret,
                                                      grant_type="authorization_code")
        response = await self.sess.post(path=uri)
        if response.status_code != 200:
            raise OAuth2Error(response, response.json())

        # construct the token
        token = OAuth2Token.from_dict(response.json())
        return token

    async def refresh_token(self, token: typing.Union[OAuth2Token, str],
                            scopes: typing.List[OAuth2Scope] = None) -> OAuth2Token:
        """
        Refreshes a token.
        
        :param token: Either a :class:`.OAuth2Token` or the str refresh token.
        :param scopes: The scopes to request.
            If an OAuth2Token is passed as the token, this will be fetched automatically.
        :return: The refreshed :class:`.OAuth2Token`.
        """
        if isinstance(token, OAuth2Token):
            ref = token.refresh_token
            scopes = [scope.name.lower() for scope in token.scopes]
        else:
            ref = token
            scopes = [scope.name.lower() for scope in scopes] if scopes else None

        uri, headers, body = self._oauth2_client.prepare_refresh_token_request(
            self.TOKEN_URL, refresh_token=ref, scope=scopes
        )

        response = await self.sess.post(path=uri, headers=list(headers.items()), body=body)

        self._oauth2_client.parse_request_body_response(body=await response.text())
        token = OAuth2Token.from_dict(await response.json())
        return token



# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
Defines :class:`.State`.

.. currentmodule:: curious.core.state
"""

import collections
import copy
import logging
import typing
from types import MappingProxyType

import multio

from curious.core import gateway
from curious.dataclasses.bases import allow_external_makes
from curious.dataclasses.channel import Channel, ChannelType
from curious.dataclasses.emoji import Emoji
from curious.dataclasses.guild import ContentFilterLevel, Guild, MFALevel, NotificationLevel, \
    VerificationLevel
from curious.dataclasses.member import Member
from curious.dataclasses.message import Message
from curious.dataclasses.permissions import Permissions
from curious.dataclasses.presence import Status
from curious.dataclasses.reaction import Reaction
from curious.dataclasses.role import Role
from curious.dataclasses.user import BotUser, User
from curious.dataclasses.voice_state import VoiceState
from curious.dataclasses.webhook import Webhook

UserType = typing.TypeVar("U", bound=User)
logger = logging.getLogger("curious.state")


def int_or_none(val, default: typing.Any) -> typing.Union[int, None]:
    """
    Returns int or None.
    """
    if val is None:
        return default

    return int(val)


class GuildStore(collections.MutableMapping):
    """
    A store for guilds in the state.
    """

    def __init__(self):
        #: The internal actual guilds.
        self.guilds = {}

        #: The order of the guilds, as specified by the READY packet.
        self.order = []

    def view(self) -> typing.Mapping[int, Guild]:
        """
        :return: A :class:`mappingproxy` of the internal guilds. 
        """
        return MappingProxyType(self.guilds)

    @property
    def with_order(self) -> 'typing.Mapping[int, Guild]':
        """
        :return: A mapping of the guilds with the order specified in the ready packet.
        """
        if not self.order:
            return self.view()

        o = collections.OrderedDict()
        for guild in map(int, self.order):
            o[guild] = self.guilds[guild]

        return MappingProxyType(o)

    # abc methods
    def __setitem__(self, key, value) -> None:
        return self.guilds.__setitem__(key, value)

    def __getitem__(self, key) -> Guild:
        return self.guilds.__getitem__(key)

    def __delitem__(self, key) -> None:
        return self.guilds.__delitem__(key)

    def __iter__(self) -> typing.Iterator[Guild]:
        return self.guilds.__iter__()

    def __len__(self) -> int:
        return self.guilds.__len__()


class State(object):
    """
    This represents the state of the Client - in other libraries, the cache.

    The other main purpose for this class is to parse events from the Discord websocket.
    """

    def __init__(self, client, max_messages: int = 500):
        #: The current user of this bot.
        #: This is automatically set after login.
        self._user = None  # type: BotUser

        #: The client associated with this connection.
        self.client = client

        #: The private channel cache.
        self._private_channels = {}

        #: The guilds the bot can see.
        self._guilds = GuildStore()

        #: The current user cache.
        self._users = {}

        #: The deque of messages.
        #: This is bounded to prevent the message cache from growing infinitely.
        self.messages = collections.deque(maxlen=max_messages)

        self.__shards_is_ready = collections.defaultdict(lambda: False)
        self.__voice_state_crap = collections.defaultdict(
            lambda *args, **kwargs: ((multio.Event(), multio.Event()), {})
        )

    def is_ready(self, shard_id: int) -> bool:
        """
        Checks if a shard is ready.
        
        :param shard_id: The shard ID to check.
        :return: A boolean signifying if this shard is ready or not.
        """
        return self.__shards_is_ready[shard_id]

    def _reset(self, shard_id: int):
        """
        Called after session is invalidated, to reset our state.
        """
        self.__shards_is_ready.pop(shard_id, None)

        for guild in self.guilds_for_shard(shard_id):
            guild._finished_chunking.clear()

    @property
    def guilds(self) -> typing.Mapping[int, Guild]:
        """
        :return: A mapping of int -> :class:`.Guild`.
        """
        return self._guilds.view()

    @property
    def guilds_ordered(self) -> typing.Mapping[int, Guild]:
        """
        :return: An ordered mapping of int -> :class:`.Guild` by the user's guild ordering.
        """
        return self._guilds.with_order

    def have_all_chunks(self, shard_id: int):
        """
        Checks if we have all the chunks for the specified shard.

        :param shard_id: The shard ID to check.
        """
        if any(guild.unavailable is True for guild in self.guilds.values()):
            return False

        return all(guild._finished_chunking.is_set()
                   for guild in self.guilds.values()
                   if guild.shard_id == shard_id and guild.unavailable is False)

    def guilds_for_shard(self, shard_id: int):
        """
        Gets all the guilds for a particular shard.
        """
        return [guild for guild in self.guilds.values() if guild.shard_id == shard_id]

    # get_all_* methods
    def get_all_channels(self) -> typing.Generator[Channel, None, None]:
        """
        :return: A generator that yields all :class:`.Channel`s the bot can see.
        """
        for guild in self._guilds.values():
            for channel in guild.channels.values():
                yield channel

    def get_all_members(self) -> typing.Generator[Member, None, None]:
        """
        :return: A generator that yields all :class:`.Member`s the bot can see.
        """
        for guild in self.guilds.values():
            for member in guild.members.values():
                yield member

    def get_all_roles(self) -> typing.Generator[Role, None, None]:
        """
        :return: A generator that yields all :class:`.Role`s the bot can see.
        """
        for guild in self.guilds.values():
            for role in guild.roles.values():
                yield role

    def find_member_or_user(self, user_id: int) -> typing.Union[Member, User]:
        """
        Finds a member or user by ID.

        :param user_id: The user ID to find.
        :return: The :class:`.Member` or :class:`.User` found, if any.
        """
        for guild in self.guilds.values():
            try:
                return guild.members[user_id]
            except KeyError:
                continue

        return self._users.get(user_id)

    def find_channel(self, channel_id: int) -> typing.Union[Channel, None]:
        """
        Finds a channel by ID.  
        This will search all guild channels, as well as private channels.
        
        :param channel_id: The ID of the channel to find.
        :return: A :class:`.Channel` that represents the channel, or None if no channel was found.
        """
        # default channel_id == guild id
        # for old guilds with a default channel
        if channel_id in self._guilds:
            try:
                return self._guilds[channel_id].channels[channel_id]
            except KeyError:
                return None

        if channel_id in self._private_channels:
            return self._private_channels[channel_id]

        for guild in self._guilds.values():
            if channel_id in guild._channels:
                return guild._channels[channel_id]

    def find_message(self, message_id: int) -> Message:
        """
        Finds a message in the current cache, if it exists.

        :param message_id: The message ID to find.
        :return: A :class:`.Message` to find, or None if it was not cached.
        """
        for message in reversed(self.messages):
            if message.id == message_id:
                return message

    def _check_decache_user(self, id: int):
        """
        Checks if we should decache a user.

        This will check if there is any guild with a reference to the user.
        """
        # don't check if its not there
        if id not in self._users:
            return

        # don't decache ourself
        if self._users[id] == self._user:
            return

        # check if its in a private channel
        for channel in self._private_channels.values():
            if id in channel.recipients:
                return

        # check if it's any guilds
        for guild in self._guilds.values():
            if id in guild.members:
                return

        # didn't return, so no references
        self._users.pop(id, None)

    # make_ methods
    def make_webhook(self, event_data: dict) -> Webhook:
        """
        Creates a new webhook object from the event data.

        :param event_data: The event data.
        :return: A :class:`.Webhook`.
        """
        if "content" in event_data:
            # message object, so we have to do a minor bit of remapping
            user = event_data.get("author", {})
            webhook_id = int(event_data.pop("webhook_id", 0))
            owner = {}
        else:
            # make a "fake" user
            webhook_id = int(event_data.get("id", 0))
            user = {
                "id": webhook_id,
                "discriminator": "0000",
                "avatar": event_data.get("avatar"),
                "username": event_data.get("username")
            }
            owner = event_data.get("user", {})

        channel = self.find_channel(int(event_data.get("channel_id")))
        user = self.make_user(user)
        # ensure the webhook user is decached
        self._check_decache_user(user.id)
        user.bot = True
        webhook = Webhook(client=self.client, webhook_id=webhook_id, **event_data)
        webhook.guild_id = channel.guild_id
        webhook.channel_id = channel.id
        webhook.user = user
        webhook.token = event_data.get("token", None)

        if owner:
            # only create Owner if the data was returned
            webhook.owner = self.make_user(owner)
            self._check_decache_user(webhook.owner.id)

        # default fields, these are lazily loaded by properties
        webhook.default_name = event_data.get("name", None)
        webhook._default_avatar = event_data.get("avatar", None)

        return webhook

    def make_private_channel(self, channel_data: dict) -> Channel:
        """
        Creates a new private channel and caches it.

        :param channel_data: The channel data to cache.
        :return: A new :class:`.Channel`.
        """
        channel = Channel(self.client, **channel_data)
        self._private_channels[channel.id] = channel

        return channel

    def make_user(self, user_data: dict, *,
                  user_klass: typing.Type[UserType] = User,
                  override_cache: bool = False) -> UserType:
        """
        Creates a new user and caches it.

        :param user_data: The user data to use to create.
        :param user_klass: The type of user to create.
        :param override_cache: Should the cache be overridden?
        :return: A new :class`.User` (hopefully).
        """
        id = int(user_data.get("id", 0))
        if id in self._users and not override_cache:
            return self._users[id]

        user = user_klass(self.client, **user_data)
        self._users[user.id] = user

        return user

    def make_message(self, event_data: dict, cache: bool = True) -> Message:
        """
        Constructs a new message object.
        
        :param event_data: The message data to use to create.
        :param cache: Should this message be cached?
        :return: A new :class:`.Message` object for the message.
        """
        message = Message(self.client, **event_data)

        if message in self.messages:
            # don't bother re-caching
            i = self.messages.index(message)
            return self.messages[i]

        # discord won't give us the Guild id
        # so we have to search it from the channels
        channel_id = int(event_data.get("channel_id", 0))
        channel = self.find_channel(channel_id)

        author_id = int(event_data.get("author", {}).get("id", 0))

        if channel is not None:
            message.guild_id = channel.guild_id
        if message.channel.type == ChannelType.PRIVATE:
            if author_id == self._user.id:
                message.author = self._user
            else:
                message.author = message.channel.user
        elif message.channel.type == ChannelType.GROUP:
            message.author = message.channel.recipients.get(author_id, None)
        else:
            # Webhooks also exist.
            if event_data.get("webhook_id") is not None:
                message.author = self.make_webhook(event_data)
            else:
                message.author = message.guild.members.get(author_id)

        for reaction_data in event_data.get("reactions", []):
            emoji = reaction_data.get("emoji", {})
            reaction = Reaction(**reaction_data)

            if "id" in emoji and emoji["id"] is not None:
                emoji_obb = message.guild.emojis.get(int(emoji["id"]))
                if emoji_obb is None:
                    emoji_obb = Emoji(id=emoji["id"], name=emoji["name"])
            else:
                emoji_obb = emoji.get("name", None)

            reaction.emoji = emoji_obb
            message.reactions.append(reaction)

        if cache and message not in self.messages:
            self.messages.append(message)

        return message

    async def wait_for_voice_data(self, guild_id: int):
        """
        Waits for the two voice data packets to be received for the specified guild.
        """
        events, state = self.__voice_state_crap[guild_id]  # state is a pointer

        for event in events:
            await event.wait()

        # pop out the event data for laeter reconnects
        self.__voice_state_crap.pop(guild_id)

        return state

    # ==============================================================================================
    # Event handlers.
    # These parse the events and deconstruct them.

    async def handle_ready(self, gw: 'gateway.GatewayHandler', event_data: dict):
        """
        Called when READY is dispatched.
        """
        # Create our bot user.
        self._user = BotUser(self.client, **event_data.get("user"))
        # cache ourselves
        self._users[self._user.id] = self._user

        logger.info("We have been issued a session on shard {}, parsing ready for `{}#{}` ({})"
                    .format(gw.gw_state.shard_id, self._user.username, self._user.discriminator,
                            self._user.id)
                    )

        # Create all of the guilds.
        for guild in event_data.get("guilds", []):
            new_guild = Guild(self.client, **guild)
            new_guild.shard_id = gw.gw_state.shard_id
            self._guilds[new_guild.id] = new_guild

        logger.info("Ready processed for shard {}. Delaying until all guilds are chunked."
                    .format(gw.gw_state.shard_id))
        yield "connect",

        # event_data.pop("guilds")
        # pprint.pprint(event_data)

    async def handle_resumed(self, gw: 'gateway.GatewayHandler', event_data: dict):
        """
        Called when the gateway connection is resumed.
        """
        yield ("resumed",)

    async def handle_user_update(self, gw: 'gateway.GatewayHandler', event_data: dict):
        """
        Called when the bot's user is updated.
        """
        id = event_data.get("id")

        self._user.id = int(id)
        self._user.username = event_data.get("username", self._user.username)
        self._user.discriminator = event_data.get("discriminator", self._user.discriminator)
        self._user.avatar_hash = event_data.get("avatar", self._user.avatar_hash)

        yield "user_update",

    async def handle_presence_update(self, gw: 'gateway.GatewayHandler', event_data: dict):
        """
        Called when a member changes game.
        """
        guild_id = int(event_data.get("guild_id", 0))
        guild = self._guilds.get(guild_id)

        # awful payloads
        user = event_data.get("user")
        if not user:
            return

        # also awful payloads
        try:
            user_id = int(event_data.get("id"))
        except (ValueError, TypeError):
            return

        # deprecated - previously friend updates, but userbot support was removed
        if not guild:
            return

        # try and create a new member from the presence update
        member = guild.members.get(user_id)
        if member is None:
            # create the member from the presence
            # we only pass the User here as we're about to update everything
            member = Member(client=self.client, user=event_data["user"])
            old_member = None
        else:
            old_member = member._copy()

        # Update the member's presence
        member.presence.status = event_data.get("status")
        member.presence.game = event_data.get("game", {})

        # copy the roles if it exists
        roles = event_data.get("roles", [])
        if roles:
            # clear roles
            member.role_ids = [int(rid) for rid in roles]

        # update the nickname
        member.nickname = event_data.get("nick", member.nickname)
        # recreate the user object, so the user is properly cached
        if "username" in event_data["user"]:
            self.make_user(event_data["user"], override_cache=True)

        yield "member_update", old_member, member,

    async def handle_presences_replace(self, gw: 'gateway.GatewayHandler', event_data: dict):
        # TODO
        print("P_R", event_data)

    async def handle_guild_members_chunk(self, gw: 'gateway.GatewayHandler', event_data: dict):
        """
        Called when a chunk of members has arrived.
        """
        id = int(event_data.get("guild_id"))
        guild = self._guilds.get(id)

        if not guild:
            logger.warning("Got a chunk for a Guild that doesn't exist...")
            return

        members = event_data.get("members", [])
        logger.info("Got a chunk of {} members in guild {} "
                    "on shard {}".format(len(members), guild.name or guild.id, guild.shard_id))

        guild._handle_member_chunk(event_data.get("members"))
        yield "guild_chunk", guild, len(members),

        if guild._chunks_left <= 0:
            # Set the finished chunking event.
            await guild._finished_chunking.set()

    async def handle_guild_create(self, gw: 'gateway.GatewayHandler', event_data: dict):
        """
        Called when GUILD_CREATE is dispatched.
        """
        id = int(event_data.get("id", 0))
        guild = self._guilds.get(id)

        had_guild = True
        if guild:
            guild.from_guild_create(**event_data)
        else:
            had_guild = False
            guild = Guild(self.client, **event_data)
            self._guilds[guild.id] = guild

        guild.shard_id = gw.gw_state.shard_id
        # TODO: Need to do this
        # try:
        #    guild.me.presence.game = gw.game
        #    guild.me.presence.status = gw.status
        # except AttributeError:
        #    # unavailable guilds etc
        #    pass

        # Dispatch the event if we're ready (i.e not streaming)
        if self.__shards_is_ready[gw.gw_state.shard_id]:
            if had_guild:
                yield "guild_available", guild,
            else:
                # We didn't have it before, so we just joined it.
                # Hence, we fire a `guild_join` event.
                # Parse the guild.
                guild.from_guild_create(**event_data)
                yield "guild_join", guild,

                logger.info("Joined guild {} ({}), requesting members if applicable"
                            .format(guild.name, guild.id))
                # if guild.large:
                #    await gw.request_chunks([guild])

        else:
            logger.debug("Streamed guild: {} ({})".format(guild.name, guild.id))
            yield "guild_streamed", guild,

        members = len(event_data.get("members", []))
        yield "guild_chunk", guild, members

    async def handle_guild_update(self, gw: 'gateway.GatewayHandler', event_data: dict):
        """
        Called when GUILD_UPDATE is dispatched.
        """
        id = int(event_data.get("id", 0))
        guild = self._guilds.get(id)

        if not guild:
            return

        # disable dataclass checking temporarily
        with allow_external_makes():
            old_guild = copy.copy(guild)

        guild.unavailable = event_data.get("unavailable", False)
        guild.name = event_data.get("name", guild.name)
        guild.member_count = event_data.get("member_count", guild.member_count)
        if not guild.member_count:
            guild.member_count = len(guild._members)
        guild._large = event_data.get("large", guild._large)
        guild.icon_hash = event_data.get("icon", guild.icon_hash)
        guild.splash_hash = event_data.get("splash", guild.splash_hash)
        guild.region = event_data.get("region", guild.region)
        guild.features = event_data.get("features", guild.features)

        guild.mfa_level = MFALevel(event_data.get("mfa_level", guild.mfa_level))
        guild.verification_level = VerificationLevel(event_data.get("verification_level",
                                                                    guild.verification_level))
        guild.notification_level = NotificationLevel(
            event_data.get("default_message_notifications", guild.notification_level)
        )
        guild.content_filter_level = ContentFilterLevel(
            event_data.get("explicit_content_filter", guild.content_filter_level)
        )

        guild.system_channel_id = int_or_none(event_data.get("system_channel_id"),
                                              guild.system_channel_id)

        guild.afk_channel_id = int_or_none(event_data.get("afk_channel"), guild.afk_channel_id)
        guild.afk_timeout = event_data.get("afk_timeout", guild.afk_timeout)
        guild.owner_id = int_or_none(event_data.get("owner_id"), guild.owner_id)

        yield "guild_update", old_guild, guild,

    async def handle_guild_delete(self, gw: 'gateway.GatewayHandler', event_data: dict):
        """
        Called when a guild becomes unavailable.
        """
        guild_id = int(event_data.get("id", 0))
        # Check if the `unavailable` flag is there.
        # If it is, we want to semi-discard this event, because all it means is the guild
        # becomes unavailable.
        if event_data.get("unavailable", False):
            # Set the guild to unavailable, but don't delete it.
            guild = self._guilds.get(guild_id)
            if guild:
                guild.unavailable = True
                yield "guild_unavailable", guild,

        else:
            # We've left this guild - clear it from our dictionary of guilds.
            guild = self._guilds.pop(guild_id, None)
            if guild:
                yield "guild_leave", guild,
                for member in guild._members.values():
                    # use member.id to avoid user lookup
                    self._check_decache_user(member.id)

    async def handle_guild_emojis_update(self, gw: 'gateway.GatewayHandler', event_data: dict):
        """
        Called when a guild updates its emojis.
        """
        guild_id = int(event_data.get("guild_id", 0))
        guild = self._guilds.get(guild_id)

        if not guild:
            return

        old_guild = guild._copy()
        emojis = event_data.get("emojis", [])
        guild._handle_emojis(emojis)

        yield "guild_emojis_update", old_guild, guild,

    async def handle_message_create(self, gw: 'gateway.GatewayHandler', event_data: dict):
        """
        Called when MESSAGE_CREATE is dispatched.
        """
        message = self.make_message(event_data)
        if not message:
            return

        # Hope that messages are ordered!
        message.channel._last_message_id = message.id

        if self._user in message.mentions:
            yield "message_mentioned", message,

        yield "message_create", message,

    async def handle_message_update(self, gw: 'gateway.GatewayHandler', event_data: dict):
        """
        Called when MESSAGE_UPDATE is dispatched.
        """
        new_message = self.make_message(event_data, cache=False)
        if not new_message:
            return

        yield "message_update_uncached", new_message

        # Try and find the old message.
        old_message = self.find_message(new_message.id)
        if not old_message:
            return

        self.messages.remove(old_message)
        self.messages.append(new_message)

        if old_message.content != new_message.content:
            # Fire a message_edit, as well as a message_update, because the content differs.
            yield "message_edit", old_message, new_message,

        yield "message_update", old_message, new_message,

    async def handle_message_delete(self, gw: 'gateway.GatewayHandler', event_data: dict):
        """
        Called when MESSAGE_DELETE is dispatched.
        """
        message_id = int(event_data.get("id"))
        yield "message_delete_uncached", message_id

        message = self.find_message(message_id)

        if not message:
            return

        yield "message_delete", message,

    async def handle_message_delete_bulk(self, gw: 'gateway.GatewayHandler', event_data: dict):
        """
        Called when MESSAGE_DELETE_BULK is dispatched.
        """
        messages = []
        ids = event_data.get("ids", [])
        yield "message_delete_bulk_uncached", ids

        for message in ids:
            message = self.find_message(int(message))
            if not message:
                continue

            messages.append(message)

        yield "message_delete_bulk", messages,

    def _find_emoji(self, emoji_data: dict):
        if emoji_data.get("id", None) is None:
            # str only
            return emoji_data["name"]

        # try and get it from the guilds
        for guild in self.guilds.values():
            em = guild.emojis.get(int(emoji_data["id"]))
            if em:
                return em

    async def handle_message_reaction_add(self, gw: 'gateway.GatewayHandler', event_data: dict):
        """
        Called when a reaction is added to a message.
        """
        message_id = event_data.get("message_id")
        if not message_id:
            return

        message_id = int(message_id)
        message = self.find_message(message_id)

        if not message:
            return

        # complex filter
        def _f(r: Reaction):
            if not r.emoji:
                return False

            e = self._find_emoji(event_data["emoji"])
            if not e:
                # ¯\_(ツ)_/¯
                return False

            return r.emoji == e

        reaction = next(filter(_f, message.reactions), None)

        if not reaction:
            emoji = event_data.get("emoji", {})
            # no useful args are added
            reaction = Reaction()

            if "id" in emoji and emoji["id"] is not None:
                emoji_obb = message.guild.emojis.get(int(emoji["id"]))
                if emoji_obb is None:
                    emoji_obb = Emoji(id=emoji["id"], name=emoji["name"])
            else:
                emoji_obb = emoji.get("name", None)

            reaction.emoji = emoji_obb
            message.reactions.append(reaction)
        else:
            # up the count
            reaction.count += 1
            # if our user id matches, we obviously voted on it
            if int(event_data["user_id"]) == self._user.id:
                reaction.me = True

        member_id = int(event_data.get("user_id", 0))
        channel = self.find_channel(int(event_data.get("channel_id", 0)))
        if channel.guild:
            author = channel.guild.members.get(member_id)
        else:
            author = channel.user

        yield "message_reaction_add", message, author, reaction,

    async def handle_message_reaction_remove_all(self, gw: 'gateway.GatewayHandler',
                                                 event_data: dict):
        """
        Called when all reactions are removed from a message.
        """
        message = self.find_message(int(event_data.get("message_id", 0)))
        if not message:
            return

        reactions = message.reactions.copy()
        message.reactions = []
        yield "message_reaction_remove_all", message, reactions,

    async def handle_message_reaction_remove(self, gw: 'gateway.GatewayHandler', event_data: dict):
        """
        Called when a reaction is removed from a message.
        """
        message_id = event_data.get("message_id")
        if not message_id:
            return

        message_id = int(message_id)
        message = self.find_message(message_id)

        if not message:
            return

        def _f(r: Reaction):
            if not r.emoji:
                return False

            e = self._find_emoji(event_data["emoji"])
            if not e:
                # ¯\_(ツ)_/¯
                return False

        reaction = next(filter(_f, message.reactions), None)
        if not reaction:
            # nothing to do
            return

        reaction.count -= 1
        if int(event_data["user_id"]) == self._user.id:
            reaction.me = False

        if reaction.count == 0:
            message.reactions.remove(reaction)

        yield "message_reaction_remove", message, reaction,

    async def handle_guild_member_add(self, gw: 'gateway.GatewayHandler', event_data: dict):
        """
        Called when a guild adds a new member.
        """
        guild_id = int(event_data.get("guild_id", 0))
        guild = self._guilds.get(guild_id)

        if not guild:
            return

        member = Member(self.client, **event_data)
        member.guild_id = guild.id

        guild._members[member.id] = member
        guild.member_count += 1
        yield "guild_member_add", member,

    async def handle_guild_member_remove(self, gw: 'gateway.GatewayHandler', event_data: dict):
        """
        Called when a guild removes a member.
        """
        guild_id = int(event_data.get("guild_id", 0))
        guild = self._guilds.get(guild_id)

        if not guild:
            return

        member = guild._members.pop(int(event_data["user"]["id"]), None)
        guild.member_count -= 1
        if not member:
            # We can't see the member, so don't fire an event for it.
            return

        yield "guild_member_remove", member,

    async def handle_guild_member_update(self, gw: 'gateway.GatewayHandler', event_data: dict):
        """
        Called when a guild member is updated.
        """
        guild_id = int(event_data.get("guild_id", 0))
        guild = self._guilds.get(guild_id)

        if not guild:
            return

        member_id = int(event_data["user"]["id"])
        member = guild.members.get(member_id)

        if not member:
            return

        # Make a copy of the member for the old previous reference.
        old_member = member._copy()
        # Re-create the user object.
        # self.make_user(event_data["user"], override_cache=True)
        # self._users[member.user.id] = member.user

        # Overwrite roles, we want to get rid of any roles that are stale.
        if "roles" in event_data:
            member.role_ids = [int(i) for i in event_data.get("roles", [])]

        guild._members[member.id] = member
        member.nickname = event_data.get("nick", member.nickname)

        yield "guild_member_update", old_member, member,

    async def handle_guild_ban_add(self, gw: 'gateway.GatewayHandler', event_data: dict):
        """
        Called when a ban is added to a guild.
        """
        guild_id = int(event_data.get("guild_id", 0))
        guild = self._guilds.get(guild_id)

        if guild is None:
            return

        member_id = int(event_data["user"]["id"])
        member = guild.members.get(member_id)

        if not member:
            # Dispatch to `user_ban` instead of `member_ban`.
            user = self.make_user(event_data["user"])
            yield "user_ban", guild, user,
            return

        yield "guild_member_ban", member,

    async def handle_guild_ban_remove(self, gw: 'gateway.GatewayHandler', event_data: dict):
        """
        Called when a ban is removed from a guild.
        """
        guild_id = int(event_data.get("guild_id", 0))
        guild = self._guilds.get(guild_id)

        if guild is None:
            return

        user = self.make_user(event_data["user"])
        yield "user_unban", guild, user,

    async def handle_channel_create(self, gw: 'gateway.GatewayHandler', event_data: dict):
        """
        Called when a channel is created.
        """
        guild_id = int(event_data.get("guild_id", 0))
        guild = self._guilds.get(guild_id)

        channel = Channel(self.client, **event_data)
        if channel.private:
            self._private_channels[channel.id] = channel
        else:
            channel.guild_id = guild.id
            channel._update_overwrites((event_data.get("permission_overwrites", [])))
            if channel.id not in guild._channels:
                guild._channels[channel.id] = channel
            else:
                channel = guild._channels[channel.id]

        yield "channel_create", channel,

    async def handle_channel_update(self, gw: 'gateway.GatewayHandler', event_data: dict):
        """
        Called when a channel is updated.
        """
        channel_id = int(event_data.get("id"))
        channel = self.find_channel(channel_id)

        if not channel:
            return

        old_channel = channel._copy()

        channel.name = event_data.get("name", channel.name)
        channel.position = event_data.get("position", channel.position)
        channel.topic = event_data.get("topic", channel.topic)
        channel.nsfw = event_data.get("nsfw", channel.nsfw)
        channel.icon_hash = event_data.get("icon_hash", channel.icon_hash)
        channel.owner_id = int_or_none(event_data.get("owner_id"), channel.owner_id)
        channel.parent_id = int_or_none(event_data.get("parent_id"), channel.parent_id)

        channel._update_overwrites(event_data.get("permission_overwrites", []))
        yield "channel_update", old_channel, channel,

    async def handle_channel_delete(self, gw: 'gateway.GatewayHandler', event_data: dict):
        """
        Called when a channel is deleted.
        """
        channel_id = int(event_data.get("channel_id", 0))
        channel = self.find_channel(channel_id)

        if not channel:
            return

        if channel.private:
            del self._private_channels[channel.id]
        else:
            del channel.guild._channels[channel.id]

        yield "channel_delete", channel,

    async def handle_guild_role_create(self, gw: 'gateway.GatewayHandler', event_data: dict):
        """
        Called when a role is created.
        """
        guild_id = int(event_data.get("guild_id", 0))
        guild = self._guilds.get(guild_id)

        if not guild:
            return

        role_data = event_data.get("role")  # type: dict
        if role_data is None:
            return

        role_id = int(role_data.get("id", 0))
        if not role_id:
            return

        if role_id not in guild._roles:
            role = Role(self.client, **role_data)
            role.guild_id = guild.id
            guild._roles[role_id] = role
        else:
            # thinking
            role = guild._roles[role_id]

        yield "role_create", role

    async def handle_guild_role_update(self, gw: 'gateway.GatewayHandler', event_data: dict):
        """
        Called when a role is updated.
        """
        guild_id = int(event_data.get("guild_id", 0))
        guild = self._guilds.get(guild_id)

        if not guild:
            return

        role_data = event_data.get("role")  # type: dict
        if role_data is None:
            return

        role_id = int(role_data.get("id", 0))
        if not role_id:
            return

        role = guild.roles.get(role_id)
        if not role:
            return

        old_role = role._copy()

        # Update all the fields on the role.
        event_data = event_data.get("role", {})
        role.colour = event_data.get("color", 0)
        role.name = event_data.get("name")
        role.position = event_data.get("position")
        role.hoisted = event_data.get("hoisted")
        role.mentionable = event_data.get("mentionable")
        role.managed = event_data.get("managed")
        role.permissions = Permissions(event_data.get("permissions", 0))

        yield "role_update", old_role, role,

    async def handle_guild_role_delete(self, gw: 'gateway.GatewayHandler', event_data: dict):
        """
        Called when a role is deleted.
        """
        guild_id = int(event_data.get("guild_id", 0))
        guild = self._guilds.get(guild_id)

        if not guild:
            return

        role = guild._roles.pop(int(event_data["role_id"]), None)
        if not role:
            return

        # Remove the role from all members.
        for member in guild.members.values():
            try:
                member.role_ids.remove(role.id)
            except ValueError:
                continue

        yield "role_delete", role,

    async def handle_typing_start(self, gw: 'gateway.GatewayHandler', event_data: dict):
        """
        Called when a user starts typing.
        """
        user_id = int(event_data.get("user_id"))
        channel_id = int(event_data.get("channel_id"))

        channel = self.find_channel(channel_id)
        if not channel:
            return

        if not channel.private:
            member = channel.guild.members.get(user_id)
            if not member:
                return
            yield "guild_member_typing", channel, member,
        else:
            user = channel.recipients.get(user_id)
            if user is None:
                return

            yield "user_typing", channel, user,

    # Voice bullshit
    async def handle_voice_server_update(self, gw: 'gateway.GatewayHandler', event_data: dict):
        """
        Called when a voice server update packet is dispatched.
        """
        # This is an internal event, so we do NOT dispatch it.
        # Instead, store some state internally.
        guild_id = event_data.get("guild_id")
        if not guild_id:
            # thanks
            return

        events, state = self.__voice_state_crap[int(guild_id)]

        state.update({
            "token": event_data.get("token"),
            "endpoint": event_data.get("endpoint"),
        })

        # Set the VOICE_SERVER_UPDATE event.
        await events[0].set()

    async def handle_voice_state_update(self, gw: 'gateway.GatewayHandler', event_data: dict):
        """
        Called when a member's voice state changes.
        """
        guild_id = int(event_data.get("guild_id", 0))

        # If user_id == self._user.id, it's voice conn bullshit
        user_id = int(event_data.get("user_id", 0))
        if user_id == self._user.id:
            # YAY
            events, state = self.__voice_state_crap[guild_id]
            state.update({
                "session_id": event_data.get("session_id")
            })
            # 2nd one is voice_state_update event
            await events[1].set()
            return

        # get the guild and member
        guild = self._guilds.get(guild_id)

        if not guild:
            return

        member = guild.members.get(user_id)
        if not member:
            return

        channel_id = event_data.get("channel_id", 0)
        if channel_id is None:
            # disconnect!
            new_voice_state = None
        else:
            new_voice_state = VoiceState(**event_data)
            new_voice_state.guild_id = guild.id

        # copy the voice states
        old_voice_state = guild._voice_states.pop(user_id, None)
        if new_voice_state is not None:
            guild._voice_states[new_voice_state.user_id] = new_voice_state

        yield "voice_state_update", member, old_voice_state, new_voice_state,

    async def handle_webhooks_update(self, gw: 'gateway.GatewayHandler', event_data: dict):
        """
        Called when a channel has a webhook updated.

        This event is effectively useless.
        """

    # TODO: Flesh these out
    async def handle_channel_pins_update(self, gw: 'gateway.GatewayHandler', event_data: dict):
        pass

    async def handle_channel_pins_ack(self, gw: 'gateway.GatewayHandler', event_data: dict):
        pass

    # Userbot only events.
    async def handle_user_settings_update(self, gw: 'gateway.GatewayHandler', event_data: dict):
        """
        Called when the current user's settings update.
        """
        old_settings = self._user.settings.copy()

        dict.update(self._user.settings, **event_data)
        # make sure to update the guild order
        guild_order = event_data.get("guild_positions")
        if guild_order:
            self._guilds.order = [int(x) for x in guild_order]

        # update status, if applicable
        new_status = Status(self._user.settings.get("status", old_settings.get("status", "ONLINE")))
        for guild in self.guilds.values():
            if new_status.strength > guild.me.status.strength:
                guild.me.status = new_status

        yield "user_settings_update", old_settings, self._user.settings,

    async def handle_message_ack(self, gw: 'gateway.GatewayHandler', event_data: dict):
        """
        Called when a message is acknowledged.
        """
        channel = self.find_channel(int(event_data.get("channel_id", 0)))
        try:
            message = self.find_message(int(event_data.get("message_id", 0)))
        except (ValueError, TypeError):
            # message_id is None, wtf?
            return

        if channel is None:
            return

        if message is None:
            return

        yield "message_ack", channel, message,

    async def handle_channel_recipient_add(self, gw: 'gateway.GatewayHandler', event_data: dict):
        """
        Called when a recipient is added to a channel.
        """
        user = event_data.get("user", {})
        id = int(event_data.get("channel_id", 0))

        user = self._users.get(int(user.get("id", 0)))

        channel = self.find_channel(channel_id=id)
        if channel is None:
            return

        channel._recipients[user.id] = user

        yield "group_user_add", channel, user,

    async def handle_channel_recipient_remove(self, gw: 'gateway.GatewayHandler', event_data: dict):
        """
        Called when a recipient is removed a channel.
        """
        user = event_data.get("user", {})
        id = int(event_data.get("channel_id", 0))

        user = self._users.get(int(user.get("id", 0)))
        channel = self.find_channel(channel_id=id)
        if channel is None:
            return

        if user in channel.recipients.values():
            channel._recipients.pop(user.id, None)
            yield "group_user_remove", channel, user,



# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
Base classes that all dataclasses inherit from.

.. currentmodule:: curious.dataclasses.bases
"""
import datetime
import inspect
import sys
import threading
from contextlib import contextmanager

from curious.core import client

DISCORD_EPOCH = 1420070400000

_allowing_external_makes = threading.local()
_allowing_external_makes.flag = False


@contextmanager
def allow_external_makes() -> None:
    """
    Using this with a ``with`` allows dataclasses to be made outside of curious' internal code.
    """
    try:
        _allowing_external_makes.flag = True
        yield
    finally:
        _allowing_external_makes.flag = False


class IDObject(object):
    """
    This object is comparable using the snowflake as an ID.

    It is also hashable, using the ID as a hash.
    """

    __slots__ = "id",

    def __init__(self, id: int):
        """
        :param id: The snowflake ID of the object.
        """
        if isinstance(id, str):
            id = int(id)

        #: The ID of this object.
        self.id = id

    def __repr__(self) -> str:
        return "<{} id={!r}>".format(self.__class__.__name__, self.id)

    __str__ = __repr__

    @property
    def snowflake_timestamp(self) -> datetime.datetime:
        """
        :return: The timestamp of the snowflake.
        """
        return datetime.datetime.utcfromtimestamp(((int(self.id) >> 22) + DISCORD_EPOCH) / 1000)

    def __eq__(self, other) -> bool:
        if not hasattr(other, "id"):
            return NotImplemented

        return other.id == self.id

    def __hash__(self) -> int:
        return hash(self.id)


class Dataclass(IDObject):
    """
    The base class for all dataclasses.

    These contain a reference to the current bot as `_bot`.
    """

    # __weakref__ is used to allow weakreffing
    __slots__ = "_bot", "__weakref__"

    @staticmethod
    def __new__(cls, *args, **kwargs):
        """
        Inspects the stack to ensure we're being called correctly.
        """
        if _allowing_external_makes.flag is False:
            try:
                frameinfo = inspect.stack()[1]
                frame = frameinfo.frame
                f_globals = frame.f_globals
                f_name = frame.f_code.co_name
                module = f_globals.get('__name__', None)
                file = f_globals.get('__file__', None)

                if module is not None:
                    if f_name == "_convert" and module.startswith("curious.commands"):
                        raise RuntimeError("You passed a dataclass ({}) as a type hint to your "
                                           "command without a converter - don't do this!\n"
                                           "This error has been raised because no builtin converter"
                                           " exists, or the built-in converter has been replaced. "
                                           "Make sure to either add one or fix your code to use a "
                                           "converter function!".format(cls.__name__))
                    elif not module.startswith("curious") \
                            and f'/python3.{sys.version_info[1]}' not in file:
                        raise RuntimeError("You tried to make a dataclass manually - don't do this!"
                                           "\nThe library handles making dataclasses for you. If "
                                           "you want to get an instance, use the appropriate "
                                           "lookup method. \nIf you really need to make the "
                                           "dataclass yourself, wrap it in a "
                                           "``with allow_external_makes)``.")
            finally:
                del frameinfo, frame

        return object.__new__(cls)

    def __init__(self, id: int, cl: 'client.Client'):
        super().__init__(id)

        self._bot = cl



# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
Wrappers for Permission objects.

This class uses some automatic generation to create the objects.

.. currentmodule:: curious.dataclasses.permissions
"""
import typing

from curious.dataclasses import channel as dt_channel, member as dt_member, role as dt_role
from curious.exc import PermissionsError


# I'm far too lazy to type out each permission bit manually.
# So here's a helper method.
def build_permissions_class(name: str = "Permissions") -> type:
    """
    Builds the permissions class automagically.
    This should ***not*** be used by normal user code - it is designed for internal usage by 
    curious.

    :param name: The name of the class.
    :return: A new type representing the permissions class.
    """
    # Closure methods.
    __doc__ = """
    Represents the permissions a user can have.
    This type is automatically generated based upon a set of constant permission bits.

    Every permission is accessible via a property getter and setter. The raw permissions value is 
    accessible via ``bitfield``.
    """

    def __init__(self, value: int = 0,
                 **kwargs):
        """
        Creates a new Permissions object.

        :param value: The bitfield value of the permissions object.
        """
        self.bitfield = value
        for perm, value in kwargs.items():
            if perm not in permissions:
                raise ValueError("Unknown permission", perm)

            setattr(self, perm, value)

    def __new__(cls, value: int = 0,
                **kwargs):
        if isinstance(value, cls):
            return value

        return super(Permissions, cls).__new__(cls)

    def _get_bit(self, bit: int) -> bool:
        """
        Gets a bit from the internal bitfield of the permissions.
        """

        return bool((self.bitfield >> bit) & 1)

    def _set_bit(self, bit: int, value: bool):
        if value:
            self.bitfield |= (1 << bit)
        else:
            self.bitfield &= ~(1 << bit)

    # Operator overloads.
    def __eq__(self, other):
        return self.bitfield == other.bitfield

    # This is a dict because discord skips some permissions.
    permissions = {
        "create_instant_invite": 0,
        "kick_members": 1,
        "ban_members": 2,
        "administrator": 3,
        "manage_channels": 4,
        "manage_server": 5,
        "add_reactions": 6,
        "view_audit_log": 7,
        "read_messages": 10,
        "send_messages": 11,
        "send_tts_messages": 12,
        "manage_messages": 13,
        "embed_links": 14,
        "attach_files": 15,
        "read_message_history": 16,
        "mention_everyone": 17,
        "use_external_emojis": 18,
        "voice_connect": 20,
        "voice_speak": 21,
        "voice_mute_members": 22,
        "voice_deafen_members": 23,
        "voice_move_members": 24,
        "voice_use_voice_activation": 25,
        "change_nickname": 26,
        "manage_nicknames": 27,
        "manage_roles": 28,
        "manage_webhooks": 29,
        "manage_emojis": 30,
        # rest are unused
    }

    # Create a bunch of property objects for each permission.
    def _get_permission_getter(name: str, bit: int):
        def _junk_function(self) -> bool:
            return self._get_bit(bit)

        _junk_function.__name__ = name
        return _junk_function

    def _get_permission_setter(name: str, bit: int):
        def _junk_function(self, value: bool):
            return self._set_bit(bit, value)

        _junk_function.__name__ = name
        return _junk_function

    _doc_base = ":return: If this member has the {} permission (bit {})."

    properties = {
        name: property(fget=_get_permission_getter(name, bit),
                       fset=_get_permission_setter(name, bit),
                       doc=_doc_base.format(name, bit)) for (name, bit) in permissions.items()
    }

    def raise_for_permission(self, permission: typing.Union[str]) -> None:
        """
        Raises :class:`.PermissionsError` if this permission does not have the required bit.
        """
        if not getattr(self, permission):
            raise PermissionsError(permission)

    # Create some useful classmethods.
    @classmethod
    def all(cls):
        """
        :return: A new Permissions object with all permissions.
        """
        return cls(9007199254740991)

    @classmethod
    def none(cls):
        """
        :return: A new permissions object with no permissions.
        """
        return cls(0)

    # Create the namespace dict to use in the type declaration.
    namespace = {
        "__init__": __init__,
        "__new__": __new__,
        "_set_bit": _set_bit,
        "_get_bit": _get_bit,
        "__eq__": __eq__,
        "__repr__": lambda self: "<Permissions value={}>".format(self.bitfield),
        "all": all,
        "none": none,
        "raise_for_permission": raise_for_permission,
        "__slots__": ("bitfield",),
        **properties
    }
    new_class = type(name, (object,), namespace)
    new_class.__doc__ = __doc__

    return new_class


Permissions = build_permissions_class("Permissions")


class Overwrite(object):
    """
    Represents a permission overwrite.

    This has all properties that the base Permissions object, but it takes into accounts the 
    overwrites for the channels. It is always recommended to use this over the server permissions, 
    as it will fall back to the default permissions for the role if it can't find specific 
    overwrites.

    The overwrite has a permission marked as ``True`` if the object has a) an overwrite on the 
    channel OR b) the object has that permission and no overwrite. The overwrite is marked as 
    ``False`` if the object has a) an overwrite on the channel OR b) the object does not have that 
    permission and no overwrite/a deny overwrite.

    You can set an attribute to None to clear the overwrite, True to set an allow overwrite, and 
    False to set a deny overwrite.
    """

    __slots__ = "target", "channel_id", "allow", "deny"

    def __init__(self, allow: typing.Union[int, Permissions], deny: typing.Union[int, Permissions],
                 obb: 'typing.Union[dt_member.Member = None, dt_role.Role]',
                 channel_id: int = None):
        """
        :param allow: A :class:`.Permissions` that this overwrite allows.
        :param deny: A :class:`.Permissions` that this overwrite denies.
        :param obb: Optional: The :class:`.Member` or :class:`.Role` that this overwrite is for.
        :param channel_id: Optional: The channel ID this overwrite is in.
        """
        self.target = obb
        self.channel_id = channel_id

        if isinstance(allow, Permissions):
            allow = allow.bitfield
        self.allow = Permissions(value=allow if allow is not None else 0)

        if isinstance(deny, Permissions):
            deny = deny.bitfield
        self.deny = Permissions(value=deny if deny is not None else 0)

    @property
    def channel(self) -> 'typing.Union[dt_channel.Channel, None]':
        """
        :return: The :class:`.Channel` this overwrite represents.
        """

    def __repr__(self) -> str:
        return "<Overwrites for object={} channel={} allow={} deny={}>".format(self.target,
                                                                               self.channel,
                                                                               self.allow,
                                                                               self.deny)

    def __getattr__(self, item) -> bool:
        """
        Attribute getter helper.

        This will check allow first, the deny, then finally the role permissions.
        """
        if isinstance(self.target, dt_member.Member):
            permissions = self.target.guild_permissions
        elif isinstance(self.target, dt_role.Role):
            permissions = self.target.permissions
        else:
            raise TypeError("Target must be a member or a role")

        if permissions.administrator:
            # short-circuit to always return True if they have administrator
            # this is because those overrides are useless
            # if the user wants to get the override, they can access `allow/deny` directly.
            return True

        if not hasattr(self.allow, item):
            raise AttributeError(item)

        if getattr(self.allow, item, None) is True:
            return True

        if getattr(self.deny, item, None) is True:
            # Return False because it's denied.
            return False

        return getattr(permissions, item, False)

    def __setattr__(self, key: str, value: object) -> object:
        """
        Attribute setter helper.
        """
        if not hasattr(Permissions, key):
            super().__setattr__(key, value)
            return

        if value is False:
            setattr(self.deny, key, True)
        elif value is True:
            setattr(self.allow, key, True)
        elif value is None:
            setattr(self.allow, key, False)
            setattr(self.deny, key, False)



# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
Wrappers for Application Info objects.

.. currentmodule:: curious.dataclasses.appinfo
"""
from typing import Union

from curious.dataclasses import user as dt_user
from curious.dataclasses.bases import Dataclass


class AppInfo(Dataclass):
    """
    Represents the application info for an OAuth2 application.
    """

    def __init__(self, client, **kwargs) -> None:
        self._application = kwargs.get("application", {})

        #: The client ID of this application.
        self.client_id = int(self._application.get("id", 0))
        super().__init__(self.client_id, client)

        if "owner" in self._application:
            owner = self._bot.state.make_user(self._application.get("owner"))
        else:
            owner = None

        #: The owner of this application.
        #: This can be None if the application fetched isn't the bot's.
        self.owner = owner

        #: The name of this application.
        self.name = self._application.get("name", None)  # type: str

        #: The description of this application.
        self.description = self._application.get("description", None)  # type: str

        #: Is this bot public?
        self.public = self._application.get("bot_public", None)  # type: bool

        #: Does this bot require OAuth2 Code Grant?
        self.requires_code_grant = self._application.get("bot_require_code_grant",
                                                         None)  # type: bool

        #: The icon hash for this application.
        self._icon_hash = self._application.get("icon", None)  # type: str

        #: The bot :class:`.User` associated with this application, if available.
        self.bot = None  # type: dt_user.User

        if "bot" in kwargs:
            self.bot = self._bot.state.make_user(kwargs.get("bot", {}))
        else:
            self.bot = None

    def __repr__(self) -> str:
        return "<{} owner='{!r}' name='{!r}' bot='{!r}'>".format(type(self).__name__, self.owner,
                                                                 self.name, self.bot)

    @property
    def icon_url(self) -> Union[str, None]:
        """
        :return: The icon url for this bot.
        """
        if self._icon_hash is None:
            return None

        return "https://cdn.discordapp.com/app-icons/{}/{}.jpg".format(self.client_id,
                                                                       self._icon_hash)




# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
Websocket wrapper classes, using a different backend library.
"""
import abc
from collections import AsyncIterable


class BasicWebsocketWrapper(AsyncIterable):
    """
    The base class for a basic websocket wrapper.
    """
    _done = object()

    def __init__(self, url: str) -> None:
        #: The gateway URL.
        self.url = url

    @abc.abstractclassmethod
    async def open(cls, url: str) -> 'BasicWebsocketWrapper':
        """
        Opens this websocket.
        """

    @abc.abstractmethod
    async def close(self, code: int = 1000, reason: str = "Client closed connection",
                    reconnect: bool = False) -> None:
        """
        Cancels and closes this websocket.

        :param code: The close code for this websocket.
        :param reason: The close reason for this websocket.
        :param reconnect: If the websocket should reconnect after being closed.
        """

    @abc.abstractmethod
    async def send_text(self, text: str) -> None:
        """
        Sends text down the websocket.
        """


# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
A trio websocket wrapper.
"""
import functools
import threading
from collections import AsyncIterator

import trio
from lomond import WebSocket
from lomond.events import Event
from lomond.persist import persist

from curious import USER_AGENT
from curious.core._ws_wrapper import BasicWebsocketWrapper


class TrioWebsocketWrapper(BasicWebsocketWrapper):
    """
    Implements a websocket handler for Trio.
    """

    def __init__(self, url: str, nursery):
        """
        :param url: The gateway URL.
        :param nursery: The nursery to use.
        """
        super().__init__(url)

        self.nursery = nursery

        self._portal = trio.BlockingTrioPortal()
        self._queue = trio.Queue(capacity=5)
        self._cancelled = threading.Event()
        self._ws = None  # type: WebSocket

    def websocket_task(self) -> None:
        """
        Runs the websocket task.
        """
        ws = WebSocket(self.url, agent=USER_AGENT)
        self._ws = ws
        # generate poll events every 0.5 seconds to see if we can cancel
        websocket = persist(ws, ping_rate=0, poll=1, exit_event=self._cancelled)
        for event in websocket:
            self._portal.run(self._queue.put, event)

        # signal the end of the queue
        self._portal.run(self._queue.put, self._done)

    @classmethod
    async def open(cls, url: str, nursery) -> 'BasicWebsocketWrapper':
        """
        Opens a new websocket connection.

        :param url: The URL to use.
        :param nursery: The nursery to use.
        """
        obb = cls(url, nursery)
        partial = functools.partial(trio.run_sync_in_worker_thread)
        nursery.start_soon(partial, obb.websocket_task)
        return obb

    async def close(self, code: int = 1000, reason: str = "Client closed connection",
                    reconnect: bool = False) -> None:
        """
        Closes the websocket.

        :param code: The close code to use.
        :param reason: The close reason to use.
        :param reconnect: If the websocket should reconnect.
        """
        if not reconnect:
            self._cancelled.set()

        self._ws.close(code=code, reason=reason)

    async def send_text(self, text: str) -> None:
        """
        Sends text down the websocket.

        :param text: The text to send.
        """
        self._ws.send_text(text)

    async def __aiter__(self) -> 'AsyncIterator[Event]':
        async for item in self._queue:
            if item == self._done:
                return

            yield item



# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
Curious - An async Python 3.6+ library for Discord bots.

.. currentmodule:: curious

.. autosummary::
    :toctree: 
    
    core
    commands
    dataclasses
    ext.paginator
    
    exc
    util
"""
from __future__ import generator_stop  # enforce generator stop

import sys

from pkg_resources import DistributionNotFound, get_distribution

try:
    __version__ = get_distribution("discord-curious").version
except DistributionNotFound:
    __version__ = "0.0.0"

_fmt = "DiscordBot (https://github.com/SunDwarf/curious {0}) Python/{1[0]}.{1[1]}"
USER_AGENT = _fmt.format(__version__, sys.version_info)
del _fmt


from curious.core.client import BotType, Client
from curious.core.event import EventContext, event
from curious.core.gateway import open_websocket, GatewayHandler
from curious.core.state import GuildStore, State
from curious.dataclasses.appinfo import AppInfo
from curious.dataclasses.attachment import Attachment
from curious.dataclasses.bases import Dataclass, IDObject
from curious.dataclasses.channel import Channel, ChannelType
from curious.dataclasses.embed import Embed
from curious.dataclasses.emoji import Emoji
from curious.dataclasses.guild import ContentFilterLevel, Guild, GuildChannelWrapper, \
    GuildEmojiWrapper, GuildRoleWrapper, MFALevel, NotificationLevel, VerificationLevel
from curious.dataclasses.invite import Invite, InviteChannel, InviteGuild, InviteMetadata
from curious.dataclasses.member import Member
from curious.dataclasses.message import Message
from curious.dataclasses.presence import Game, Presence, Status
from curious.dataclasses.reaction import Reaction
from curious.dataclasses.role import Role
from curious.dataclasses.search import SearchQuery, SearchResults
from curious.dataclasses.user import User
from curious.dataclasses.voice_state import VoiceState
from curious.dataclasses.webhook import Webhook
from curious.dataclasses.widget import Widget, WidgetChannel, WidgetGuild, WidgetMember

# for asks
# import multio
# multio.init('curio')



# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
The main Discord HTTP interface.

.. currentmodule:: curious.core.httpclient
"""
import datetime
import logging
import mimetypes
import random
import time
import typing
import weakref
from email.utils import parsedate
from math import ceil, floor
from urllib.parse import quote

import asks
import multio
import pytz
from asks.errors import ConnectivityError
from asks.response_objects import Response
from h11 import RemoteProtocolError

try:
    # try and load a C impl of LRU first
    from lru import LRU as c_lru

    lru = c_lru
except ImportError:
    # fall back to a pure-python (the default) version
    from pylru import lrucache as py_lru

    lru = py_lru

import curious
from curious.exc import Forbidden, HTTPException, NotFound, Unauthorized

logger = logging.getLogger("curious.http")


def parse_date_header(header: str) -> datetime.datetime:
    """
    Parses a date header.

    :param header: The contents of the header to parse.
    :return: A :class:`datetime.datetime` that corresponds to the date header.
    """
    dt = datetime.datetime(*parsedate(header)[:6], tzinfo=pytz.UTC)
    return dt


def encode_multipart(fields, files, boundary=None):
    r"""Encode dict of form fields and dict of files as multipart/form-data.
    Return tuple of (body_string, headers_dict). Each value in files is a dict
    with required keys 'filename' and 'content', and optional 'mimetype' (if
    not specified, tries to guess mime type or uses 'application/octet-stream').
    >>> body, headers = encode_multipart({'FIELD': 'VALUE'},
    ...                                  {'FILE': {'filename': 'F.TXT', 'content': 'CONTENT'}},
    ...                                  boundary='BOUNDARY')
    >>> print('\n'.join(repr(l) for l in body.split('\r\n')))
    '--BOUNDARY'
    'Content-Disposition: form-data; name="FIELD"'
    ''
    'VALUE'
    '--BOUNDARY'
    'Content-Disposition: form-data; name="FILE"; filename="F.TXT"'
    'Content-Type: text/plain'
    ''
    'CONTENT'
    '--BOUNDARY--'
    ''
    >>> print(sorted(headers.items()))
    [('Content-Length', '193'), ('Content-Type', 'multipart/form-data; boundary=BOUNDARY')]
    >>> len(body)
    193
    Copied from: https://code.activestate.com/recipes/578668-encode-multipart-form-data-for-uploading-files-via/
    """
    _BOUNDARY_CHARS = "oanfwenfuwengvwenvnewuifn34fy8u0newfwer69420"

    def escape_quote(s):
        return s.replace(b'"', b'\\"')

    if boundary is None:
        boundary = b''.join(random.choice(_BOUNDARY_CHARS).encode() for i in range(30))
    lines = []

    for name, value in fields.items():
        if isinstance(name, str):
            name = name.encode()
        else:
            name = str(name).encode()

        if isinstance(value, str):
            value = value.encode()
        else:
            value = str(value).encode()

        lines.extend((
            b'--%s' % boundary,
            b'Content-Disposition: form-data; name="%s"' % escape_quote(name),
            b'',
            value,
        ))

    for name, value in files.items():
        filename = value['filename']
        if 'mimetype' in value:
            mimetype = value['mimetype']
        else:
            mimetype = mimetypes.guess_type(filename)[0] or 'application/octet-stream'
        lines.extend((
            b'--%s' % boundary,
            b'Content-Disposition: form-data; name="%s"; filename="%s"' % (
                escape_quote(name.encode()), escape_quote(filename.encode())),
            b'Content-Type: %s' % mimetype.encode(),
            b'',
            value['content'],
        ))

    lines.extend((
        b'--%s--' % boundary,
        b'',
    ))
    body = b'\r\n'.join(lines)

    headers = {
        'Content-Type': 'multipart/form-data; boundary=%s' % boundary.decode(),
        'Content-Length': str(len(body)),
    }

    return body, headers


# more of a namespace
class Endpoints:
    API_BASE = "/api/v7"

    USER_ID = "/users/{user_id}"
    USER_ME = "/users/@me"
    USER_MENTIONS = USER_ME + "/mentions"
    USER_GUILD = USER_ME + "/guilds/{guild_id}"
    USER_CHANNELS = USER_ME + "/channels"

    GATEWAY = "/gateway"
    GATEWAY_BOT = "/gateway/bot"

    GUILD_BASE = "/guilds"
    GUILD_ID_BASE = "/guilds/{guild_id}"
    GUILD_CHANNELS = GUILD_ID_BASE + "/channels"
    GUILD_MEMBERS = GUILD_ID_BASE + "/members"
    GUILD_MEMBER = GUILD_MEMBERS + "/{member_id}"
    GUILD_MEMBER_NICK_ME = GUILD_MEMBERS + "/@me/nick"
    GUILD_MEMBER_ROLE = GUILD_MEMBER + "/{role_id}"
    GUILD_VANITY_URL = GUILD_ID_BASE + "/vanity-url"
    GUILD_BANS = GUILD_ID_BASE + "/bans"
    GUILD_BAN_USER = GUILD_BANS + "/{user_id}"
    GUILD_ROLES = GUILD_ID_BASE + "/roles"
    GUILD_ROLE = GUILD_ROLES + "/{role_id}"
    GUILD_WIDGET = GUILD_ID_BASE + "/widget.json"
    GUILD_EMBED = GUILD_ID_BASE + "/embed"
    GUILD_AUDITLOGS = GUILD_ID_BASE + "/audit-logs"
    GUILD_WEBHOOKS = GUILD_ID_BASE + "/webhooks"
    GUILD_INVITES = GUILD_ID_BASE + "/invites"
    GUILD_EMOJIS = GUILD_ID_BASE + "/emojis"
    GUILD_EMOJI = GUILD_EMOJIS + "/{emoji_id}"

    CHANNEL_BASE = "/channels/{channel_id}"
    CHANNEL_TYPING = CHANNEL_BASE + "/typing"
    CHANNEL_MESSAGES = CHANNEL_BASE + "/messages"
    CHANNEL_MESSAGES_SEARCH = CHANNEL_MESSAGES + "/search"
    CHANNEL_MESSAGE = CHANNEL_MESSAGES + "/{message_id}"
    CHANNEL_MESSAGE_REACTIONS = CHANNEL_MESSAGE + "/reactions"
    CHANNEL_MESSAGE_REACTION_EMOJI = CHANNEL_MESSAGE_REACTIONS + "/{emoji}"
    CHANNEL_MESSAGE_REACTION_ME = CHANNEL_MESSAGE_REACTION_EMOJI + "/@me"
    CHANNEL_MESSAGE_REACTION_OTHER = CHANNEL_MESSAGE_REACTION_EMOJI + "/{victim}"
    CHANNEL_MESSAGE_BULK_DELETE = CHANNEL_MESSAGES + "/bulk-delete"
    CHANNEL_PINS = CHANNEL_BASE + "/pins"
    CHANNEL_PIN_MESSAGE = CHANNEL_PINS + "/{message_id}"
    CHANNEL_PERMISSION = CHANNEL_BASE + "/{target_id}"
    CHANNEL_WEBHOOKS = CHANNEL_BASE + "/webhooks"
    CHANNEL_INVITES = CHANNEL_BASE + "/invites"

    WEBHOOKS_BASE = "/webhooks"
    WEBHOOKS_GET = "/webhooks/{webhook_id}"
    WEBHOOKS_TOKEN = "/webhooks/{webhook_id}/{token}"

    INVITES_BASE = "/invites"
    INVITE_GET = "/invites/{invite_code}"

    OAUTH2_BASE = "/oauth2"
    OAUTH2_AUTHORIZE = OAUTH2_BASE + "/authorize"
    OAUTH2_APPLICATIONS = OAUTH2_BASE + "/applications"
    OAUTH2_APPLICATION_ME = OAUTH2_APPLICATIONS + "/@me"
    OAUTH2_TOKENS = OAUTH2_BASE + "/tokens"
    OAUTH2_REVOKE = OAUTH2_TOKENS + "/{app_id}"

    def __init__(self, base_url: str = "https://discordapp.com"):
        """
        :param base_url: The base URL for this set of endpoints.
        """
        self.BASE = base_url


class HTTPClient(object):
    """
    The HTTP client object used to make requests to Discord's servers.

    If a particular method is not listed here, you can use one of the five following methods to make
    a manual request:

        - :meth:`HTTPClient.get`
        - :meth:`HTTPClient.post`
        - :meth:`HTTPClient.put`
        - :meth:`HTTPClient.delete`
        - :meth:`HTTPClient.patch`

    All of these functions require a **ratelimit bucket** which will be used to prevent the client 
    from hitting 429 ratelimits.

    :param token: The token to use for all HTTP requests.
    :param bot: Is this client a bot?
    :param max_connections: The max connections for this HTTP client.
    """

    def __init__(self, token: str, *,
                 bot: bool = True,
                 max_connections: int = 10):
        #: The token used for all requests.
        self.token = token

        # Calculated headers
        headers = {
            "User-Agent": curious.USER_AGENT,
            "Authorization": "{}{}".format("Bot " if bot else "", self.token)
        }

        self.endpoints = Endpoints()
        self.session = asks.Session(base_location=self.endpoints.BASE, endpoint=Endpoints.API_BASE,
                                    connections=max_connections)
        self.headers = headers

        #: The global ratelimit lock.
        self.global_lock = multio.Lock()

        self._rate_limits = weakref.WeakValueDictionary()
        self._ratelimit_remaining = lru(1024)
        self._is_bot = bot

    def get_ratelimit_lock(self, bucket: object) -> 'multio.Lock':
        """
        Gets a ratelimit lock from the dict if it exists, otherwise creates a new one.
        """
        try:
            return self._rate_limits[bucket]
        except KeyError:
            lock = multio.Lock()
            self._rate_limits[bucket] = lock
            return lock

    # Special wrapper functions
    @staticmethod
    def get_response_data(response: Response) -> typing.Union[str, dict]:
        """
        Return either the text of a request or the JSON.

        :param response: The response to use.
        """
        if response.headers.get("Content-Type", None) == "application/json":
            return response.json()

        return response.content

    async def _make_request(self, *args, **kwargs) -> Response:
        """
        Makes a request via the current session.

        :returns: The response body.
        """
        headers = kwargs.get("headers", None)
        if headers is not None:
            headers.update(self.headers.copy())
            kwargs.pop("headers")
        else:
            headers = self.headers.copy()

        # update reason header
        if "reason" in kwargs:
            headers["X-Audit-Log-Reason"] = quote(kwargs["reason"])

        # ensure path is escaped
        path = kwargs["path"]
        path = quote(path)
        kwargs["path"] = path

        # temporary
        # return await self.session.request(*args, headers=headers, timeout=5, **kwargs)
        if 'uri' not in kwargs:
            kwargs["uri"] = self.endpoints.BASE + Endpoints.API_BASE + kwargs["path"]
        else:
            kwargs.pop("path", None)

        return await asks.request(*args, headers=headers, timeout=5, **kwargs)

    async def request(self, bucket: object, *args, **kwargs):
        """
        Makes a rate-limited request.

        This will respect Discord's X-Ratelimit-Limit headers to make requests.

        :param bucket: The bucket this request falls under.
        """
        # Okay, an English explaination of how this works.
        # First, it loads the curio-based lock from the defaultdict of lock, keyed by bucket.
        # Then, it tries to acquire the lock. This is held by one request at a time, naturally.

        # Normally, the lock is immediately released upon a request finishing, which allows the
        # next request to handle it. However, once X-Ratelimit-Remaining is 0, we don't want any
        # more requests to be made until the time limit is over. So the request sleeps for
        # (X-RateLimit-Reset - time.time()) seconds, then unlocks the lock.

        lock = self.get_ratelimit_lock(bucket)
        # If we're being globally ratelimited, this will block until the global lock is finished.
        await self.global_lock.acquire()
        # Immediately release it because we're no longer being globally ratelimited.
        await self.global_lock.release()
        try:
            await lock.acquire()

            if bucket in self._ratelimit_remaining:
                # Make sure we have enough tries left.
                tries, reset_time = self._ratelimit_remaining[bucket]

                if tries <= 0:
                    # We need to sleep for a bit before we can start making another request.
                    sleep_time = ceil(reset_time - time.time())
                    logger.debug("Sleeping with lock open for {} seconds.".format(sleep_time))
                    await multio.asynclib.sleep(sleep_time)

            for tries in range(0, 5):
                method = kwargs.get("method", "???")
                path = kwargs.get("path", "???")
                logger.debug(f"{method} {path} => (pending) (try {tries + 1})")

                try:
                    response = await self._make_request(*args, **kwargs)
                except OSError:
                    # discord forcefully disconnected or similar
                    continue
                except ConnectivityError:
                    # discord deadlocked for whatever reason
                    continue
                except RemoteProtocolError:
                    # discord broke
                    continue

                logger.debug(f"{method} {path} => {response.status_code} (try {tries + 1})")

                if response.status_code in range(500, 600):
                    # 502 means that we can retry without worrying about ratelimits.
                    # Perform exponential backoff to prevent spamming discord.
                    sleep_time = 1 + (tries * 2)
                    await multio.asynclib.sleep(sleep_time)
                    continue

                if response.status_code == 429:
                    # This is bad!
                    # But it's okay, we can handle it.
                    logger.warning("Hit a 429 in bucket {}. Check your clock!".format(bucket))
                    sleep_time = ceil(int(response.headers["Retry-After"]) / 1000)
                    await multio.asynclib.sleep(sleep_time)
                    continue

                # Extract ratelimit headers.
                remaining = int(response.headers.get("X-Ratelimit-Remaining", 1))
                reset = int(response.headers.get("X-Ratelimit-Reset", 1))

                # Update the ratelimit headers.
                self._ratelimit_remaining[bucket] = remaining, reset

                # Next, check if we need to sleep.
                # Check if we need to sleep.
                # This is signaled by Ratelimit-Remaining being 0 or Ratelimit-Global being True.
                is_global = response.headers.get("X-Ratelimit-Global", None) is not None
                should_sleep = remaining == 0 or is_global

                if should_sleep:
                    # The time until the reset is given by X-Ratelimit-Reset.
                    # Failing that, it's also given by the Retry-After header, which is in ms.
                    reset = response.headers.get("X-Ratelimit-Reset")
                    # Parse Discord's Date header to use their time rather than local time.
                    parsed_time = parse_date_header(response.headers.get("Date")).timestamp()
                    if reset:
                        sleep_time = int(reset) - parsed_time
                    else:
                        retry_after = response.headers.get("Retry-After")
                        if retry_after is not None:
                            sleep_time = ceil(int(response.headers.get("Retry-After")) / 1000)
                        else:
                            # fallback in case we get some really bad response
                            sleep_time = 1 + (tries * 2)

                    before_time = time.monotonic()
                    if is_global:
                        logger.debug("Reached the global ratelimit, acquiring global lock.")
                        await self.global_lock.acquire()
                    after_time = int(floor(time.monotonic() - before_time))
                    if after_time != 0:
                        # subtract the time we spent waiting for the global lock to be acquired
                        sleep_time -= after_time

                    logger.debug(
                        "Being ratelimited under bucket %s, waking in %s seconds",
                        bucket, sleep_time
                    )

                    # Sleep that amount of time.
                    await multio.asynclib.sleep(sleep_time)
                    # If the global lock is acquired, unlock it now
                    if is_global:
                        await self.global_lock.release()

                # Now, we have that nuisance out of the way, we can try and get the result from
                # the request.
                result = self.get_response_data(response)

                # Status codes between 200 and 300 mean success, so we return the data directly.
                if 200 <= response.status_code < 300:
                    return result

                # Status codes between 400 and 600 are BAD!
                # So we raise an exception.
                # However, special case 404 and 403, because they're Unique Exceptions(tm).
                if 400 <= response.status_code < 600:
                    if response.status_code == 401:
                        raise Unauthorized(response, result)

                    if response.status_code == 403:
                        raise Forbidden(response, result)

                    if response.status_code == 404:
                        raise NotFound(response, result)

                    raise HTTPException(response, result)
            else:
                raise RuntimeError("Failed to get response after 5 tries.")

        finally:
            await lock.release()
            # Only release the global lock if we need to
            if self.global_lock.locked():
                await self.global_lock.release()

    async def get(self, url: str, bucket: str,
                  *args, **kwargs):
        """
        Makes a GET request.

        :param url: The URL to request.
        :param bucket: The ratelimit bucket to file this request under.
        """
        return await self.request(("GET", bucket), method="GET", path=url, *args, **kwargs)

    async def post(self, url: str, bucket: str,
                   *args, **kwargs):
        """
        Makes a POST request.

        :param url: The URL to request.
        :param bucket: The ratelimit bucket to file this request under.
        """
        return await self.request(("POST", bucket), method="POST", path=url, *args, **kwargs)

    async def put(self, url: str, bucket: str,
                  *args, **kwargs):
        """
        Makes a PUT request.

        :param url: The URL to request.
        :param bucket: The ratelimit bucket to file this request under.
        """
        return await self.request(("PUT", bucket), method="PUT", path=url, *args, **kwargs)

    async def delete(self, url: str, bucket: str,
                     *args, **kwargs):
        """
        Makes a DELETE request.

        :param url: The URL to request.
        :param bucket: The ratelimit bucket to file this request under.
        """
        return await self.request(("DELETE", bucket), method="DELETE", path=url, *args, **kwargs)

    async def patch(self, url: str, bucket: str,
                    *args, **kwargs):
        """
        Makes a PATCH request.

        :param url: The URL to request.
        :param bucket: The ratelimit bucket to file this request under.
        """
        return await self.request(("PATCH", bucket), method="PATCH", path=url, *args, **kwargs)

    # Non-generic methods
    async def get_gateway_url(self):
        """
        It is not recommended to use this method - use :meth:`HTTPClient.get_shard_count` instead. 
        That method provides the gateway URL as well.

        :return: The websocket gateway URL to get.
        """
        data = await self.get(Endpoints.GATEWAY, "gateway")
        return data["url"]

    async def get_shard_count(self):
        """
        :return: The recommended number of shards for this bot.
        """
        if not self._is_bot:
            raise Forbidden(None, {"code": 20002, "message": "Only bots can use this endpoint"})

        data = await self.get(Endpoints.GATEWAY_BOT, "gateway")
        return data["url"], data["shards"]

    async def get_this_user(self):
        """
        Gets the current user.
        """
        data = await self.get(Endpoints.USER_ME, bucket="user:get")
        return data

    async def get_user(self, user_id: int):
        """
        Gets a user from a user ID.

        :param user_id: The ID of the user to fetch.
        :return: A user dictionary.
        """
        url = Endpoints.USER_ID.format(user_id=user_id)

        # user_id isn't a major param, so handle under one bucket
        data = await self.get(url, bucket="user:get")
        return data

    async def get_guild(self, guild_id: int):
        """
        Gets a guild by guild ID.

        :param guild_id: The ID of the guild to get.
        :return: A guild object.
        """
        url = Endpoints.GUILD_BASE.format(guild_id=guild_id)

        data = await self.get(url, bucket="guild:{}".format(guild_id))
        return data

    async def get_guild_channels(self, guild_id: int):
        """
        Gets a list of channels in a guild.

        :param guild_id: The ID of the guild to get.
        :return: A list of channel objects.
        """
        url = Endpoints.GUILD_CHANNELS.format(guild_id=guild_id)

        data = await self.get(url, bucket="guild:{}".format(guild_id))
        return data

    async def get_guild_members(self, guild_id: int, *,
                                limit: int = None, after: int = None):
        """
        Gets guild members for the specified guild.

        :param guild_id: The ID of the guild to get.
        :param limit: The maximum number of members to get.
        :param after: The ID to fetch members after.
        """
        url = Endpoints.GUILD_MEMBERS.format(guild_id=guild_id)
        params = {}

        if limit is not None:
            params["limit"] = limit

        if after is not None:
            params["after"] = after

        data = await self.get(url, bucket="guild:{}".format(guild_id), params=params)
        return data

    async def get_guild_member(self, guild_id: int, member_id: int):
        """
        Gets a guild member.

        :param guild_id: The guild ID to get.
        :param member_id: The member ID to get.
        """
        url = Endpoints.GUILD_MEMBER.format(guild_id=guild_id, member_id=member_id)

        data = await self.get(url, bucket="guild:{}".format(guild_id))
        return data

    async def get_channel(self, channel_id: int):
        """
        Gets a channel.

        :param channel_id: The channel ID to get.
        """
        url = Endpoints.CHANNEL_BASE.format(channel_id=channel_id)

        data = await self.get(url, bucket="channel:{}".format(channel_id))
        return data

    async def get_vanity_url(self, guild_id: int):
        """
        Gets the vanity URL for a guild.

        :param guild_id: The guild ID to get the vanity URL of.
        """
        data = await self.get(Endpoints.GUILD_VANITY_URL, bucket="guild:{}".format(guild_id))
        return data

    async def edit_vanity_url(self, guild_id: int, code: str):
        """
        Edits the vanity URL for a guild.

        :param guild_id: The guild ID to edit the vanity URL in.
        :param code: The code of the vanity url (the part after the /).
        """
        payload = {
            "code": code,
        }

        data = await self.patch(Endpoints.GUILD_VANITY_URL, bucket="guild:{}".format(guild_id),
                                json=payload)
        return data

    async def send_typing(self, channel_id: str):
        """
        Starts typing in a channel.

        :param channel_id: The ID of the channel to type in.
        """
        url = Endpoints.CHANNEL_TYPING.format(channel_id=channel_id)

        data = await self.post(url, bucket="typing:{}".format(channel_id))
        return data

    async def send_message(self, channel_id: int, content: str, tts: bool = False,
                           embed: dict = None):
        """
        Sends a message to a channel.

        :param channel_id: The ID of the channel to send to.
        :param content: The content of the message.
        :param tts: Is this message a text to speech message?
        :param embed: The embed dict to send with this message.
        """
        url = Endpoints.CHANNEL_MESSAGES.format(channel_id=channel_id)
        payload = {
            "tts": tts,
        }

        if content is not None:
            payload["content"] = content

        if embed is not None:
            payload["embed"] = embed

        data = await self.post(url, "messages:{}".format(channel_id), json=payload)
        return data

    async def send_file(self, channel_id: int, file_content: bytes, *,
                        filename: str = None, content: str = None):
        """
        Uploads a file to the current channel.

        This will encode the data as multipart/form-data.

        :param channel_id: The channel ID to upload to.
        :param file_content: The content of the file being uploaded.
        :param filename: The filename of the file being uploaded.
        :param content: Any optional message content to send with this file.
        """
        url = Endpoints.CHANNEL_MESSAGES.format(channel_id=channel_id)
        payload = {}
        if content is not None:
            payload["content"] = content

        files = {
            "file": {
                "filename": filename,
                "content": file_content
            }
        }

        body, headers = encode_multipart(payload, files)

        data = await self.post(url, "messages:{}".format(channel_id),
                               data=body, headers=headers)
        return data

    async def delete_message(self, channel_id: int, message_id: int):
        """
        Deletes a message.

        This requires the MANAGE_MESSAGES permission.

        :param channel_id: The channel ID that the message is in.
        :param message_id: The message ID of the message.
        """
        url = Endpoints.CHANNEL_MESSAGE.format(channel_id=channel_id, message_id=message_id)

        data = await self.delete(url, "messages:{}".format(channel_id))
        return data

    async def edit_message(self, channel_id: int, message_id: int, content: str = None,
                           embed: dict = None):
        """
        Edits a message.

        This will only work on your own messages.

        :param channel_id: The channel ID that the message is in.
        :param message_id: The message ID of the message.
        :param content: The new content of the message.
        :param embed: The new embed of the message.
        """
        url = Endpoints.CHANNEL_MESSAGE.format(channel_id=channel_id, message_id=message_id)
        payload = {}

        if content is not None:
            payload["content"] = content

        if embed is not None:
            payload["embed"] = embed

        data = await self.patch(url, "messages:{}".format(channel_id), json=payload)
        return data

    async def add_reaction(self, channel_id: int, message_id: int, emoji: str):
        """
        Reacts to a message.

        :param channel_id: The channel ID that the message is in.
        :param message_id: The message ID of the message.
        :param emoji: The emoji to react with.
        """
        url = Endpoints.CHANNEL_MESSAGE_REACTION_ME.format(
            channel_id=channel_id,
            message_id=message_id,
            emoji=emoji
        )

        data = await self.put(url, "reactions:{}".format(channel_id))
        return data

    async def delete_reaction(self, channel_id: int, message_id: int, emoji: str,
                              victim: int = None):
        """
        Deletes a reaction from a message.

        :param channel_id: The channel ID of the channel containing the message.
        :param message_id: The message ID to remove reactions from.
        :param emoji: The emoji to remove.
        :param victim: The victim to remove. \
            If this is None, our own reaction is removed.
        """
        if not victim:
            ep = Endpoints.CHANNEL_MESSAGE_REACTION_ME
        else:
            ep = Endpoints.CHANNEL_MESSAGE_REACTION_OTHER
        url = ep.format(channel_id=channel_id, message_id=message_id, emoji=emoji, victim=victim)

        data = await self.delete(url, bucket="reactions:{}".format(channel_id))
        return data

    async def delete_all_reactions(self, channel_id: int, message_id: int):
        """
        Removes all reactions from a message.

        :param channel_id: The channel ID of the channel containing the message.
        :param message_id: The message ID to remove reactions from.
        """
        url = Endpoints.CHANNEL_MESSAGE_REACTIONS.format(channel_id=channel_id,
                                                         message_id=message_id)

        data = await self.delete(url, bucket="reactions:{}".format(channel_id))
        return data

    async def get_reaction_users(self, channel_id: int, message_id: int, emoji: str):
        """
        Gets a list of users who reacted to this message with the specified reaction.

        :param channel_id: The channel ID to check in.
        :param message_id: The message ID to check.
        :param emoji: The emoji to get reactions for.
        """
        url = Endpoints.CHANNEL_MESSAGE_REACTION_EMOJI.format(
            channel_id=channel_id,
            message_id=message_id,
            emoji=emoji)

        data = await self.get(url, bucket="reactions:{}".format(channel_id))
        return data

    async def pin_message(self, channel_id: int, message_id: int):
        """
        Pins a message to the channel.

        :param channel_id: The channel ID to pin in.
        :param message_id: The message ID of the message to pin.
        """
        url = Endpoints.CHANNEL_PIN_MESSAGE.format(channel_id=channel_id,
                                                   message_id=message_id)

        data = await self.put(url, "pins:{}".format(channel_id), json={})
        return data

    async def unpin_message(self, channel_id: int, message_id: int):
        """
        Unpins a message from the channel.

        :param channel_id: The channel ID to unpin in.
        :param message_id: The message ID of the message to unpin.
        """
        url = Endpoints.CHANNEL_PIN_MESSAGE.format(channel_id=channel_id,
                                                   message_id=message_id)

        data = await self.delete(url, "pins:{}".format(channel_id))
        return data

    async def get_message(self, channel_id: int, message_id: int):
        """
        Gets a single message from the channel.

        :param channel_id: The channel ID to get the message from.
        :param message_id: The message ID of the message to get.
        :return: The message data.
        """
        url = Endpoints.CHANNEL_MESSAGE.format(channel_id=channel_id, message_id=message_id)

        data = await self.get(url, "messages:{}".format(channel_id))
        return data

    async def get_message_history(self, channel_id: int, *,
                                  before: int = None, after: int = None, around: int = None,
                                  limit: int = 100):
        """
        Gets a list of messages from a channel.

        This requires READ_MESSAGES on the channel.

        :param channel_id: The channel ID to receive messages from.
        :param before: Get messages before this snowflake.
        :param after: Get messages after this snowflake.
        :param around: Get messages around this snowflake.
        :param limit: The maximum number of messages to return.
        :return: A list of message dictionaries.
        """
        url = Endpoints.CHANNEL_MESSAGES.format(channel_id=channel_id)
        payload = {
            "limit": str(limit)
        }

        if before:
            payload["before"] = str(before)

        if after:
            payload["after"] = str(after)

        if around:
            payload["around"] = str(around)

        data = await self.get(url, bucket="messages:{}".format(channel_id), params=payload)
        return data

    async def get_pins(self, channel_id: int):
        """
        Gets the pins for a channel.

        :param channel_id: The channel ID to get pins from.
        """
        url = Endpoints.CHANNEL_PINS.format(channel_id=channel_id)

        data = await self.get(url, bucket="pins:{}".format(channel_id))
        return data

    async def delete_multiple_messages(self, channel_id: int, message_ids: typing.List[int]):
        """
        Deletes multiple messages.

        This will silently discard any messages that don't exist.

        This requires MANAGE_MESSAGES on the channel, regardless of what messages are being deleted.

        :param channel_id: The channel ID to delete messages from.
        :param message_ids: A list of messages to delete.
        """
        url = Endpoints.CHANNEL_MESSAGE_BULK_DELETE.format(channel_id=channel_id)
        payload = {
            "messages": [str(message_id) for message_id in message_ids]
        }

        data = await self.post(url, bucket="messages:bulk_delete:{}".format(channel_id),
                               json=payload)
        return data

    # Profile endpoints
    async def edit_user(self, username: str = None, avatar: str = None,
                        password: str = None):
        """
        Edits the profile of the bot.

        :param username: The new username of the bot, or None if it is not to be changed.
        :param avatar: The new avatar of the bot, or None if it not to be changed.
        :param password: The password of the bot. Only needed for user bots.
        """
        url = Endpoints.USER_ME
        payload = {}
        if username:
            payload["username"] = username

        if avatar:
            payload["avatar"] = avatar

        if password:
            payload["password"] = password

        data = await self.patch(url, bucket="users:edit", json=payload)
        return data

    # Moderation
    async def get_bans(self, guild_id: int):
        """
        Gets a list of bans from a guild.

        :param guild_id: The guild to get bans from.
        :return: A list of user dicts containing ban information.
        """
        url = Endpoints.GUILD_BANS.format(guild_id=guild_id)

        data = await self.get(url, bucket="bans:{}".format(guild_id))
        return data

    async def kick_member(self, guild_id: int, member_id: int):
        """
        Kicks a member from a guild.

        :param guild_id: The guild ID to kick in.
        :param member_id: The member ID to kick from the guild.
        """
        url = Endpoints.GUILD_MEMBER.format(guild_id=guild_id,
                                            member_id=member_id)

        data = await self.delete(url, bucket="members:{}".format(guild_id))
        return data

    async def ban_user(self, guild_id: int, user_id: int,
                       delete_message_days: int = 7, reason: str = None):
        """
        Bans a user from a guild.

        :param guild_id: The ID of the guild to ban on.
        :param user_id: The user ID to ban from the guild.
        :param delete_message_days: The number of days to delete messages from this user.
        :param reason: The reason for this ban.
        """
        url = Endpoints.GUILD_BAN_USER.format(guild_id=guild_id, user_id=user_id)
        payload = {"reason": reason}

        if delete_message_days:
            payload["delete-message-days"] = delete_message_days

        data = await self.put(url, bucket="bans:{}".format(guild_id), json=payload)
        return data

    async def unban_user(self, guild_id: int, user_id: int, reason: str = None):
        """
        Unbans a user from a guild.

        :param guild_id: The ID of the guild to unban on.
        :param user_id: The user ID that has been forgiven.
        :param reason: The reason for this unban.
        """
        # TODO: Do reasons properly
        url = Endpoints.GUILD_BAN_USER.format(guild_id=guild_id, user_id=user_id)

        data = await self.delete(url, bucket="bans:{}".format(guild_id), reason=reason)
        return data

    async def create_guild(self, name: str, region: str = None, icon: str = None,
                           verification_level: int = None,
                           default_message_notifications: int = None,
                           roles: typing.List[dict] = None, channels: typing.List[dict] = None):
        """
        Creates a new guild.

        :param name: The name of the new guild.
        :param region: The region of the new guild.
        :param icon: The base64 icon of the new guild.
        :param verification_level: The verification level of the new guild.
        :param default_message_notifications: The default notification level of the new guild.
        :param roles: A dict of role objects.
        :param channels: A dict of channel objects.
        """
        url = Endpoints.GUILD_BASE
        payload = {"name": name}

        if icon:
            payload["icon"] = icon

        if region:
            payload["region"] = region

        if verification_level is not None:
            payload["verification_level"] = str(verification_level)

        if default_message_notifications is not None:
            payload["default_message_notifications"] = str(default_message_notifications)

        if roles:
            payload["roles"] = roles

        if channels:
            payload["channels"] = channels

        data = await self.post(url, "guild:create", json=payload)
        return data

    async def edit_guild(self, guild_id: int, *,
                         name: str = None, icon_content: bytes = None,
                         region: str = None, verification_level: int = None,
                         default_message_notifications: int = None,
                         afk_channel_id: int = None, afk_timeout: int = None,
                         splash_content: bytes = None,
                         explicit_content_filter: int = None,
                         system_channel_id: int = None):
        """
        Modifies a guild.

        See https://discordapp.com/developers/docs/resources/guild#modify-guild for the fields 
        available.
        """
        url = Endpoints.GUILD_BASE.format(guild_id=guild_id)
        payload = {}

        if name:
            payload["name"] = name

        if icon_content:
            payload["icon"] = icon_content

        if region:
            payload["region"] = region

        if verification_level is not None:
            payload["verification_level"] = str(verification_level)

        if default_message_notifications is not None:
            payload["default_message_notifications"] = str(default_message_notifications)

        if not afk_channel_id:
            payload["afk_channel_id"] = None
        elif afk_channel_id:
            payload["afk_channel_id"] = str(afk_channel_id)

        if afk_timeout:
            payload["afk_timeout"] = str(afk_timeout)

        if splash_content:
            payload["splash"] = splash_content

        if explicit_content_filter is not None:
            payload["explicit_content_filter"] = str(explicit_content_filter)

        if system_channel_id is not None:
            payload["system_channel_id"] = str(system_channel_id)

        data = await self.patch(url, bucket="guild_edit:{}".format(guild_id), json=payload)
        return data

    async def create_role(self, guild_id: int) -> dict:
        """
        Creates a role in a guild.

        :param guild_id: The guild to create the role in.
        """
        url = Endpoints.GUILD_ROLES.format(guild_id=guild_id)

        data = await self.post(url, bucket="guild_roles:{}".format(guild_id))
        return data

    async def edit_role(self, guild_id: int, role_id: int,
                        name: str = None, permissions: int = None, position: int = None,
                        colour: int = None, hoist: bool = None, mentionable: bool = None):
        """
        Edits a role.

        :param guild_id: The guild ID that contains the role.
        :param role_id: The role ID to edit.
        :param name: The new name of the role.
        :param permissions: The new permissions for the role.
        :param position: The position for the role.
        :param colour: The colour for the role.
        :param hoist: Is this role hoisted?
        :param mentionable: Is this role mentionable?
        """
        url = Endpoints.GUILD_ROLE.format(guild_id=guild_id, role_id=role_id)
        payload = {}

        if name:
            payload["name"] = name

        if permissions is not None:
            payload["permissions"] = permissions

        if position is not None:
            payload["position"] = position

        if colour:
            payload["color"] = colour

        if hoist is not None:
            payload["hoist"] = hoist

        if mentionable is not None:
            payload["mentionable"] = mentionable

        data = await self.patch(url, bucket="guild_roles:{}".format(guild_id), json=payload)
        return data

    async def delete_role(self, guild_id: int, role_id: int):
        """
        Deletes a role.

        :param guild_id: The guild ID that contains the role.
        :param role_id: The role ID to delete.
        """
        url = Endpoints.GUILD_ROLE.format(guild_id=guild_id, role_id=role_id)

        data = await self.delete(url, bucket="guild_roles:{}".format(guild_id))
        return data

    async def create_channel(self, guild_id: int, name: str, type: int, *,
                             bitrate: int = None, user_limit: int = None,
                             permission_overwrites: list = None):
        """
        Creates a new channel.

        :param guild_id: The guild ID to create the channel in.
        :param name: The name of the channel.
        :param type: The type of the channel (text/voice).
        :param bitrate: The bitrate of the channel, if it is a voice channel.
        :param user_limit: The maximum number of users that can be in the channel.
        :param permission_overwrites: The list of permission overwrites to use for this channel.
        """
        url = Endpoints.GUILD_CHANNELS.format(guild_id=guild_id)
        payload = {
            "name": name,
            "type": type
        }

        if type == 2:
            if bitrate is not None:
                payload["bitrate"] = bitrate

            if user_limit is not None:
                payload["user_limit"] = user_limit

        if permission_overwrites is not None:
            payload["permission_overwrites"] = permission_overwrites

        data = await self.post(url, bucket="guild_channels:{}".format(guild_id), json=payload)
        return data

    async def edit_channel(self, channel_id: int, *,
                           name: str = None, position: int = None,
                           topic: str = None,
                           bitrate: int = None, user_limit: int = -1):
        """
        Edits a channel.

        :param channel_id: The channel ID to edit.
        :param name: The new name of the channel.
        :param position: The new position of the channel.
        :param topic: The new topic of the channel.
        :param bitrate: The new bitrate of the channel.
        :param user_limit: The user limit of the channel.
        """
        url = Endpoints.CHANNEL_BASE.format(channel_id=channel_id)
        payload = {}

        if name is not None:
            payload["name"] = name

        if position is not None:
            payload["position"] = position

        if topic is not None:
            payload["topic"] = topic

        if bitrate is not None:
            payload["bitrate"] = bitrate

        if user_limit != -1:
            payload["user_limit"] = user_limit

        data = await self.patch(url, bucket="channels:{}".format(channel_id), json=payload)
        return data

    async def update_channel_positions(self, channels: typing.List[typing.Tuple[int, int]]):
        """
        Updates the positions of
        """

    async def delete_channel(self, channel_id: int):
        """
        Deletes a channel.

        :param channel_id: The channel ID to delete.
        """
        url = Endpoints.CHANNEL_BASE.format(channel_id)

        data = await self.delete(url, bucket="channels:{}".format(channel_id))
        return data

    async def add_member_role(self, guild_id: int, member_id: int, role_id: int):
        """
        Adds a single role to a member.

        If you want to add multiple roles to a member, use :meth:`add_roles`.

        :param guild_id: The guild ID that contains the objects.
        :param member_id: The member ID to add the role to.
        :param role_id: The role ID to add to the member.
        """
        url = Endpoints.GUILD_MEMBER_ROLE.format(guild_id=guild_id,
                                                 member_id=member_id,
                                                 role_id=role_id)

        data = await self.put(url, bucket="member_edit:{}".format(guild_id))
        return data

    async def edit_member_roles(self, guild_id: int, member_id: int,
                                role_ids: typing.Iterable[int]):
        """
        Modifies the roles that a member object contains.

        :param guild_id: The guild ID that contains the objects.
        :param member_id: The member ID to add the role to.
        :param role_ids: The role IDs to add to the member.
        """
        url = Endpoints.GUILD_MEMBER.format(guild_id=guild_id,
                                            member_id=member_id)
        payload = {
            "roles": [str(id) for id in role_ids]
        }

        data = await self.patch(url, bucket="member_edit:{}".format(guild_id), json=payload)
        return data

    async def edit_role_positions(self, guild_id: int,
                                  role_mapping: typing.List[typing.Tuple[str, int]]):
        """
        Changes the position of a set of roles.

        :param guild_id: The guild ID that contains the roles.
        :param role_mapping: An iterable of `(role_id, new_position)` values.
        """
        url = Endpoints.GUILD_ROLES.format(guild_id=guild_id)
        payload = [(str(r_id), position) for r_id, position in role_mapping]

        data = await self.patch(url, bucket="roles", json=payload)
        return data

    async def change_nickname(self, guild_id: int, nickname: str, *, member_id: int = None,
                              me: bool = False):
        """
        Changes the nickname of a member.

        If `me` is True, then `member_id` is not required. Otherwise, `member_id` is required.

        :param guild_id: The guild ID that contains the member.
        :param nickname: The nickname to set, None to reset.
        :param member_id: The member ID to change the nickname of.
        :param me: If this should change our own nickname.
        """
        if me:
            url = Endpoints.GUILD_MEMBER_NICK_ME.format(guild_id=guild_id)
        else:
            url = Endpoints.GUILD_MEMBER.format(guild_id=guild_id, member_id=member_id)
        payload = {
            "nick": nickname
        }

        data = await self.patch(url, bucket="member_edit:{}".format(guild_id), json=payload)
        return data

    async def edit_member_voice_state(self, guild_id: int, member_id: int, *,
                                      deaf: bool = None, mute: bool = None, channel_id: int = None):
        """
        Edits the voice state of a member.

        :param guild_id: The guild ID to edit in.
        :param member_id: The member ID to edit.
        :param deaf: Should the member be deafened?
        :param mute: Should the member be muted?
        :param channel_id: What channel should the member be moved to?
        """
        url = Endpoints.GUILD_MEMBER.format(guild_id=guild_id,
                                            member_id=member_id)
        payload = {}

        if deaf is not None:
            payload["deaf"] = deaf

        if mute is not None:
            payload["mute"] = mute

        if channel_id is not None:
            payload["channel_id"] = channel_id

        data = await self.patch(url, bucket="member_edit:{}".format(guild_id), json=payload)
        return data

    async def edit_overwrite(self, channel_id: int, target_id: int, type_: str,
                             *, allow: int = 0, deny: int = 0):
        """
        Modifies or adds an overwrite.

        :param channel_id: The channel ID to edit.
        :param target_id: The target of the override.
        :param type_: The type the target is.

        :param allow: The permission bitfield of permissions to allow.
        :param deny: The permission bitfield of permissions to deny.
        """
        url = Endpoints.CHANNEL_PERMISSION.format(channel_id=channel_id, target_id=target_id)
        payload = {
            "allow": allow,
            "deny": deny,
            "type": type_
        }

        data = await self.put(url, bucket="channels:permissions:{}".format(channel_id),
                              json=payload)
        return data

    async def remove_overwrite(self, channel_id: int, target_id: int):
        """
        Removes an overwrite.

        :param channel_id: The channel ID to edit.
        :param target_id: The target of the override.
        """
        url = Endpoints.CHANNEL_PERMISSION.format(channel_id=channel_id,
                                                  target_id=target_id)

        data = await self.delete(url, bucket="channels:permissions:{}".format(channel_id))
        return data

    async def get_widget_status(self, guild_id: int):
        """
        Gets the current widget status information for a guild.
        
        :param guild_id: The guild ID to fetch. 
        """
        url = Endpoints.GUILD_EMBED.format(guild_id=guild_id)

        data = await self.get(url, bucket="widget:{}".format(guild_id))
        return data

    async def get_widget_data(self, guild_id: int):
        """
        Gets the current widget data for a guild.
        
        :param guild_id: The guild ID of the widget to fetch.
        """
        url = Endpoints.GUILD_WIDGET.format(guild_id=guild_id)

        data = await self.get(url, bucket="widget:{}".format(guild_id))
        return data

    async def edit_widget(self, guild_id: int,
                          enabled: bool = None, channel_id: int = 0):
        """
        Edits the widget status for this guild.
        
        :param guild_id: The guild ID to edit the widget of. 
        :param enabled: Is the widget enabled in this guild? 
        :param channel_id: What channel ID is the instant invite for? This can be None to disable \ 
            the channel.
        """
        url = Endpoints.GUILD_EMBED.format(guild_id=guild_id)
        payload = {}

        if enabled is not None:
            payload["enabled"] = enabled

        if channel_id != 0:
            payload["channel_id"] = channel_id

        data = await self.patch(url, bucket="widget:{}".format(guild_id), json=payload)
        return data

    async def get_audit_logs(self, guild_id: int,
                             *, limit: int = 50, user_id: int = None,
                             action_type: int = None):
        """
        Gets the audit log for this guild.
        
        :param guild_id: The guild ID to get audit logs for.
        :param limit: The maximum number of entries to return.
        :param user_id: The user ID to filter by.
        :param action_type: The action type to filter by.
        """
        url = Endpoints.GUILD_AUDITLOGS.format(guild_id=guild_id)
        payload = {}

        if limit is not None:
            payload["limit"] = limit

        if user_id is not None:
            payload["user_id"] = user_id

        if action_type is not None:
            payload["action_type"] = action_type

        data = await self.get(url, bucket="guild:{}:audit-logs".format(guild_id), params=payload)
        return data

    # Emojis
    async def get_guild_emojis(self, guild_id: int):
        """
        Gets the emojis for a guild.

        :param guild_id: The guild ID to get emojis for.
        """
        url = Endpoints.GUILD_EMOJIS.format(guild_id=guild_id)

        data = await self.get(url, bucket=f"emojis:{guild_id}")
        return data

    async def get_guild_emoji(self, guild_id: int, emoji_id: int):
        """
        Gets an emoji object.

        :param guild_id: The guild ID the emoji is in.
        :param emoji_id: The emoji ID to get.
        """
        url = Endpoints.GUILD_EMOJI.format(guild_id=guild_id, emoji_id=emoji_id)

        data = await self.get(url, bucket=f"emojis:{guild_id}")
        return data

    async def create_guild_emoji(self, guild_id: int, *,
                                 name: str, image: str,
                                 roles: typing.List[int] = None):
        """
        Creates an emoji in a guild.

        :param guild_id: The ID of the guild to create the emoji in.
        :param name: The name of the emoji.
        :param image: The base64 image data for the emoji.
        :param roles: A list of roles this emoji is limited to.
        """
        url = Endpoints.GUILD_EMOJIS.format(guild_id=guild_id)
        params = {
            "name": name,
            "image": image,
        }
        if roles is not None:
            params["roles"] = [str(r) for r in roles]

        data = await self.post(url, bucket=f"emojis:{guild_id}", params=params)
        return data

    async def edit_guild_emoji(self, guild_id: int, emoji_id: int, *,
                               name: str = None, roles: typing.List[int] = None):
        """
        Modifies an emoji in a guild.

        :param guild_id: The ID of the guild the emoji is in.
        :param emoji_id: The ID of the emoji to modify.
        :param name: The name of the emoji to edit.
        :param roles: A list of roles this emoji is limited to.
        """
        url = Endpoints.GUILD_EMOJI.format(guild_id=guild_id, emoji_id=emoji_id)
        params = {}

        if name is not None:
            params["name"] = name

        if roles is not None:
            params["roles"] = [str(r) for r in roles]

        data = await self.patch(url, bucket=f"emojis:{guild_id}", params=params)
        return data

    async def delete_guild_emoji(self, guild_id: int, emoji_id: int):
        """
        Deletes an emoji in a guild.

        :param guild_id: The ID of the guild the emoji is in.
        :param emoji_id: The ID of the emoji to delete.
        """
        url = Endpoints.GUILD_EMOJI.format(guild_id=guild_id, emoji_id=emoji_id)

        data = await self.delete(url, bucket=f"emojis:{guild_id}")
        return data

    # Webhooks
    async def get_webhook(self, webhook_id: int):
        """
        Gets a webhook object for the specified ID.

        :param webhook_id: The ID of the webhook to get.
        """
        url = Endpoints.WEBHOOKS_GET.format(webhook_id=webhook_id)

        data = await self.get(url, bucket="webhooks")  # not a major param :(
        return data

    async def get_webhooks_for_guild(self, guild_id: int):
        """
        Gets the webhooks for the specified guild.

        :param guild_id: The ID of the guild to get the webhooks for.
        """
        url = Endpoints.GUILD_WEBHOOKS.format(guild_id=guild_id)

        data = await self.get(url, bucket="webhooks:{}".format(guild_id))
        return data

    async def get_webhooks_for_channel(self, channel_id: int):
        """
        Gets the webhooks for the specified channel.

        :param channel_id: The ID of the channel to get the webhooks for.
        """
        url = Endpoints.CHANNEL_WEBHOOKS.format(channel_id=channel_id)

        data = await self.get(url, bucket="webhooks:{}".format(channel_id))
        return data

    async def create_webhook(self, channel_id: int, *,
                             name: str = None, avatar: str = None):
        """
        Creates a webhook.

        :param channel_id: The channel ID to create the webhook in.
        :param name: The name of the webhook to create.
        :param avatar: The base64 encoded avatar to send.
        """
        url = Endpoints.CHANNEL_WEBHOOKS.format(channel_id=channel_id)
        payload = {"name": name}

        if avatar is not None:
            payload["avatar"] = avatar

        data = await self.post(url, bucket="webhooks:{}".format(channel_id), json=payload)
        return data

    async def edit_webhook(self, webhook_id: int, *,
                           name: str = None, avatar: str = None):
        """
        Edits a webhook.

        :param webhook_id: The ID of the webhook to edit.
        :param name: The name of the webhook.
        :param avatar: The base64 encoded avatar to send.
        """
        url = Endpoints.WEBHOOKS_GET.format(webhook_id=webhook_id)
        payload = {}

        if avatar is not None:
            payload["avatar"] = avatar

        if name is not None:
            payload["name"] = name

        data = await self.patch(url, bucket="webhooks", json=payload)
        return data

    async def edit_webhook_with_token(self, webhook_id: int, token: str, *,
                                      name: str = None, avatar: str = None):
        """
        Edits a webhook, with a token.

        :param webhook_id: The ID of the webhook to edit.
        :param token: The token of the webhook to edit.
        :param name: The name of the webhook to edit.
        :param avatar: The base64 encoded avatar to send.
        """
        url = Endpoints.WEBHOOKS_TOKEN.format(webhook_id=webhook_id,
                                              token=token)
        payload = {}

        if avatar is not None:
            payload["avatar"] = avatar

        if name is not None:
            payload["name"] = name

        data = await self.patch(url, bucket="webhooks", json=payload)
        return data

    async def delete_webhook(self, webhook_id: int):
        """
        Deletes a webhook.

        :param webhook_id: The ID of the webhook to delete.
        """
        url = Endpoints.WEBHOOKS_GET.format(webhook_id=webhook_id)

        data = await self.delete(url, bucket="webhooks")
        return data

    async def delete_webhook_with_token(self, webhook_id: int, token: str):
        """
        Deletes a webhook with a token.

        :param webhook_id: The ID of the webhook to delete.
        :param token: The token of the webhook.
        """
        url = Endpoints.WEBHOOKS_TOKEN.format(webhook_id=webhook_id,
                                              token=token)

        data = await self.delete(url, bucket="webhooks")
        return data

    async def execute_webhook(self, webhook_id: int, webhook_token: str, *,
                              content: str = None, embeds: typing.List[typing.Dict] = None,
                              username: str = None, avatar_url: str = None,
                              wait: bool = False):
        """
        Executes a webhook.

        :param webhook_id: The ID of the webhook to execute.
        :param webhook_token: The token of this webhook.
        :param content: Any message content to send.
        :param embeds: A list of embeds to send.
        :param username: The username to override with.
        :param avatar_url: The avatar URL to send.
        :param wait: If we should wait for the message to send.
        """
        url = Endpoints.WEBHOOKS_TOKEN.format(webhook_id=webhook_id, token=webhook_token)
        payload = {}

        if content:
            payload["content"] = content

        if embeds:
            payload["embeds"] = embeds

        if username:
            payload["username"] = username

        if avatar_url:
            payload["avatar_url"] = avatar_url

        # URL params, not payload
        params = {"wait": str(wait)}
        data = await self.post(url, bucket="webhooks", json=payload, params=params)

        return data

    # Invites
    async def get_invite(self, invite_code: str, *,
                         with_counts: bool = True):
        """
        Gets an invite by code.

        :param invite_code: The invite to get.
        :param with_counts: Should the estimated total and online members be included?
        """
        url = Endpoints.INVITES_BASE.format(invite_code=invite_code)
        params = {
            "with_counts": "true" if with_counts else "false"
        }

        data = await self.get(url, bucket="invites", params=params)
        return data

    async def get_invites_for(self, guild_id: int):
        """
        Gets the invites for the specified guild.

        :param guild_id: The guild ID to get invites inside.
        """
        url = Endpoints.GUILD_INVITES.format(guild_id=guild_id)

        data = await self.get(url, bucket="invites:{}".format(guild_id))
        return data

    async def create_invite(self, channel_id: int, *,
                            max_age: int = None, max_uses: int = None,
                            temporary: bool = None, unique: bool = None):
        """
        Creates an invite.

        :param channel_id: The channel ID to create the invite in.
        :param max_age: The maximum age of the invite.
        :param max_uses: The maximum uses of the invite.
        :param temporary: Is this invite temporary?
        :param unique: Is this invite unique?
        """
        url = Endpoints.CHANNEL_INVITES.format(channel_id=channel_id)
        payload = {}

        if max_age is not None:
            payload["max_age"] = max_age

        if max_uses is not None:
            payload["max_uses"] = max_uses

        if temporary is not None:
            payload["temporary"] = temporary

        if unique is not None:
            payload["unique"] = unique

        data = await self.post(url, bucket="invites:{}".format(channel_id), json=payload)
        return data

    async def delete_invite(self, invite_code: str):
        """
        Deletes the invite specified by the code.

        :param invite_code: The code of the invite to delete.
        """
        url = Endpoints.INVITE_GET.format(invite_code=invite_code)

        data = await self.delete(url, bucket="invites")
        return data

    async def search_channel(self, channel_id: int, params: dict):
        """
        Searches a channel.
        
        :param channel_id: The channel ID of the channel to search. 
        :param params: Params to search with.
        """
        url = Endpoints.CHANNEL_MESSAGES_SEARCH.format(channel_id=channel_id)

        data = await self.get(url, bucket="search:{}".format(channel_id), params=params)
        return data

    async def search_guild(self, guild_id: int, params: dict):
        """
        Searches a guild.

        :param guild_id: The guild ID of the guild to search. 
        :param params: Params to search with.
        """
        url = (self.GUILD_BASE + "/messages/search").format(guild_id=guild_id)

        data = await self.get(url, bucket="search:{}".format(guild_id), params=params)
        return data

    # User only
    async def update_user_settings(self, **settings):
        """
        Updates the current user's settings.
                 
        .. warning::
        
            This is a **user-acount only** endpoint.
        
        
        :param settings: The dict of settings to update. 
        """
        url = Endpoints.USER_ME + "/settings"

        data = await self.patch(url, bucket="user", json=settings)
        return data

    # Application info
    async def get_app_info(self, application_id: typing.Union[int, None]):
        """
        Gets some basic info about an application.

        :param application_id: The ID of the application to get info about.
            If this is None, it will fetch the current application info.
        """
        if application_id is None:
            return await self._get_app_info_me()

        application_id = str(application_id)

        url = Endpoints.OAUTH2_AUTHORIZE

        try:
            data = await self.get(url, bucket="oauth2",
                                  params={"client_id": application_id, "scope": "bot"})
        except HTTPException as e:
            if e.error_code != 50010:
                raise

            data = await self.get(url, bucket="oauth2", params={"client_id": application_id})
        return data

    async def _get_app_info_me(self):
        """
        :return: The application info for this bot.
        """
        url = Endpoints.OAUTH2_APPLICATION_ME

        data = await self.get(url, "oauth2")
        # httpclient is meant to be a "pure" wrapper, but add this anyway.
        me = await self.get_this_user()

        final = {
            "application": data,
            "bot": me,
        }

        return final

    async def get_user_applications(self):
        """
        Gets the list of applications for a user.
        """
        url = Endpoints.OAUTH2_APPLICATIONS

        data = await self.get(url, "oauth2")
        return data

    async def get_application(self, application_id: int):
        """
        Gets an application by ID.

        :param application_id: The ID of the application to get.
        """
        url = "/oauth2/applications/{}".format(application_id)

        data = await self.get(url, "oauth2")
        return data

    async def authorize_bot(self, application_id: int, guild_id: int,
                            *, permissions: int = 0):
        """
        Authorize a bot to be added to a guild.
        
        .. warning::
        
            This is a **user-acount only** endpoint.
        

        :param application_id: The client ID of the bot to be added.
        :param guild_id: The guild ID to add the bot to.
        :param permissions: The permissions to add the bot with.
        """
        url = Endpoints.OAUTH2_AUTHORIZE
        params = {
            "client_id": str(application_id),
            "scope": "bot"
        }

        payload = {
            "guild_id": guild_id,
            "authorize": True,
            "permissions": permissions
        }

        data = await self.post(url, "oauth2", params=params, json=payload)
        return data

    async def get_authorized_apps(self):
        """
        Gets authorized apps for this account.
                
        .. warning::
        
            This is a **user-account only** endpoint.
        
        """
        url = Endpoints.OAUTH2_TOKENS

        data = await self.get(url, bucket="oauth2")
        return data

    async def revoke_authorized_app(self, app_id: int):
        """
        Revokes an application authorization.
                
        .. warning::
        
            This is a **user-acount only** endpoint.
        
        
        :param app_id: The ID of the application to revoke the authorization of. 
        """
        url = "/oauth2/tokens/{app_id}".format(app_id=app_id)

        data = await self.delete(url, bucket="oauth")
        return data

    async def get_mentions(self, *,
                           guild_id: int = None, limit: int = 25,
                           roles: bool = True, everyone: bool = True):
        """
        Gets your recent mentions.
        
        .. warning::
        
            This is a **user-acount only** endpoint.
        
        :param guild_id: The guild ID to limit mentions for.
        :param limit: The maximum number of messages to return.
        :param roles: Should role mentions be included?
        :param everyone: Should @everyone/@here mentions be included?
        """
        url = Endpoints.USER_MENTIONS
        params = {}

        if limit is not None:
            params["limit"] = str(limit)

        if guild_id is not None:
            params["guild_id"] = str(guild_id)

        if roles is not None:
            params["roles"] = str(roles).lower()

        if everyone is not None:
            params["everyone"] = str(everyone).lower()

        data = await self.get(url, "me:mentions", params=params)
        return data

    # Misc
    async def create_private_channel(self, user_id: int):
        """
        Opens a new private channel with a user.

        :param user_id: The user ID of the user to open with.
        """
        url = Endpoints.USER_CHANNELS
        payload = {
            "recipient_id": user_id
        }

        data = await self.post(url, "channels:create", json=payload)
        return data

    async def leave_guild(self, guild_id: str):
        """
        Leaves a guild.

        :param guild_id: The guild ID of the guild to leave.
        """
        url = Endpoints.USER_GUILD.format(guild_id)

        data = await self.delete(url, "guild:leave")
        return data



# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.

"""
The core of Curious.

This package contains the bulk of the network interface with Discord, including parsing data that is incoming and 
delegating it to client code.

.. currentmodule:: curious.core

.. autosummary::
    :toctree: core
    
    client
    event
    gateway
    httpclient
    state
"""




# This file is part of curious.
#
# curious is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# curious is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with curious.  If not, see <http://www.gnu.org/licenses/>.
"""
Misc utilities used in commands related things.

.. currentmodule:: curious.commands.utils
"""
import collections
import inspect
from typing import Callable, Iterable, List, Union

import typing_inspect

from curious.commands.exc import ConversionFailedError, MissingArgumentError
from curious.core.client import Client
from curious.dataclasses.message import Message
from curious.util import replace_quotes


def get_full_name(func) -> str:
    """
    Gets the full name of a command, taking into account it's parents.
    """
    name = []
    # loop over and extract the name
    # then set func to the parent
    # and break if the parent is None
    while True:
        name.append(func.cmd_name)
        if func.cmd_parent is None:
            break

        func = func.cmd_parent

    return ' '.join(reversed(name))


async def _convert(ctx, tokens: List[str], signature: inspect.Signature):
    """
    Converts tokens passed from discord, using a signature.
    """
    final_args = []
    final_kwargs = {}

    def _with_reraise(func, ann, ctx, arg):
        try:
            return func(ann, ctx, arg)
        except ConversionFailedError:
            raise
        except Exception as e:
            raise ConversionFailedError(ctx, arg, ann, message="Converter error") from e

    args_it = iter(tokens)
    for n, (name, param) in enumerate(signature.parameters.items()):
        if n == 0:
            # Don't convert the `ctx` argument.
            continue

        assert isinstance(param, inspect.Parameter)
        # We loop over the signature parameters because it's easier to use those to consume.
        # Get the next argument from args.

        def consume_token() -> str:
            try:
                return next(args_it)
            except StopIteration as e:
                # not good!
                # If we're a *arg format, we can safely handle this, or if we have a default.
                if param.kind in [inspect.Parameter.KEYWORD_ONLY,
                                  inspect.Parameter.VAR_POSITIONAL]:
                    return None

                if param.default is inspect.Parameter.empty:
                    raise MissingArgumentError(ctx, param.name) from e

                return None  # ??

        arg = consume_token()
        if arg is None:
            break

        # Begin the consumption!
        if param.kind in [inspect.Parameter.POSITIONAL_OR_KEYWORD,
                          inspect.Parameter.POSITIONAL_ONLY]:
            # ensure we have a non-empty argument
            while arg == "":
                if arg is None:
                    break

                arg = next(args_it)

            arg = replace_quotes(arg)
            converter = ctx._lookup_converter(param.annotation)
            final_args.append(_with_reraise(converter, param.annotation, ctx, arg))
            continue

        if param.kind in [inspect.Parameter.KEYWORD_ONLY]:
            # Only add it to final_kwargs.
            # This is a consume all operation, so we eat all of the arguments.
            f = [arg]

            while True:
                next_arg = consume_token()
                if next_arg is None:
                    break

                f.append(next_arg)

            converter = ctx._lookup_converter(param.annotation)
            if len(f) == 1:
                final_kwargs[param.name] = _with_reraise(converter, param.annotation, ctx, f[0])
            else:
                final_kwargs[param.name] = _with_reraise(converter, param.annotation, ctx,
                                                         " ".join(f))
            continue

        if param.kind in [inspect.Parameter.VAR_POSITIONAL]:
            # This *shouldn't* be called on `*` arguments, but we can't be sure.
            # Special case - consume ALL the arguments.
            f = [arg]

            while True:
                next_arg = consume_token()
                if next_arg is None:
                    break

                f.append(next_arg)

            converter = ctx._lookup_converter(param.annotation)
            final_args.append(
                _with_reraise(converter, param.annotation, ctx, " ".join(f))
            )

        if param.kind in [inspect.Parameter.VAR_KEYWORD]:
            # no
            continue

    return final_args, final_kwargs


def get_description(func) -> str:
    """
    Gets the description of a function.

    :param func: The function.
    :return: The description extracted from the docstring, or None.
    """
    if not func.__doc__:
        return None

    doc = inspect.cleandoc(inspect.getdoc(func))
    lines = doc.split("\n")
    return lines[0]


def get_usage(func, invoked_as: str = None) -> str:
    """
    :return: The usage text for this command.
    """
    if invoked_as is not None:
        final = [invoked_as]
    else:
        final = [func.cmd_name]

    signature = inspect.signature(func)

    # TODO: Replace this with a proper one
    def stringify(ann):
        origin = typing_inspect.get_origin(ann)
        if not origin:
            return ann.__name__

        args = typing_inspect.get_args(ann, evaluate=True)
        return f"{origin.__name__}[{', '.join(arg.__name__ for arg in args)}]"

    for n, (name, param) in enumerate(signature.parameters.items()):
        # always skip the first arg, as it's self/ctx
        if n == 0:
            continue

        # check if we should skip the 2nd arg
        # not always possible
        from curious.commands.context import Context
        if name in ["ctx", "context"] or param.annotation is Context:
            continue

        # switch based on kind (again...)
        assert isinstance(param, inspect.Parameter)
        if param.kind in [inspect.Parameter.POSITIONAL_OR_KEYWORD,
                          inspect.Parameter.POSITIONAL_ONLY]:
            if param.annotation is not inspect.Parameter.empty:
                s = "<{}: {}>".format(param.name, stringify(param.annotation))
            else:
                s = "<{}>".format(param.name)

        elif param.kind in [inspect.Parameter.KEYWORD_ONLY, inspect.Parameter.VAR_POSITIONAL]:
            if param.default is not inspect.Parameter.empty:
                if param.annotation is not inspect.Parameter.empty:
                    s = "[{}: {} (default: {})]".format(param.name, stringify(param.annotation),
                                                        repr(param.default))
                else:
                    s = "[{} (default: {})]".format(param.name, repr(param.default))
            else:
                if param.annotation is not inspect.Parameter.empty:
                    s = "<{}: {}>".format(param.name, stringify(param.annotation))
                else:
                    s = "<{}>".format(param.name)
        else:
            s = ""

        final.append(s)

    return " ".join(final)


# This function is a modified version of the code taken from https://stackoverflow.com/a/43035638.
# This function is licenced under the MIT Licence. (C) 2017 THE_MAD_KING.
# See: https://meta.stackexchange.com/questions/272956/a-new-code-license-the-mit-this-time-with
# -attribution-required
# You can find a copy of the MIT Licence at https://opensource.org/licenses/MIT.
def split_message_content(content: str, delim: str = " ") -> List[str]:
    """
    Splits a message into individual parts by `delim`, returning a list of strings.
    This method preserves quotes.

    .. code-block:: python3

        content = '!send "Fuyukai desu" "Hello, world!"'
        split = split_message_content(content, delim=" ")

    :param content: The message content to split.
    :param delim: The delimiter to split on.
    :return: A list of items split
    """

    tokens = []
    cur = ''
    in_quotes = False

    for char in content.strip():
        if char == delim and not in_quotes:
            tokens.append(cur)
            cur = ''
        elif char == '"' and not in_quotes:
            in_quotes = True
            cur += char
        elif char == '"' and in_quotes:
            in_quotes = False
            cur += char
        else:
            cur += char
    tokens.append(cur)

    return tokens


def prefix_check_factory(prefix: Union[str, Iterable[str], Callable[[Client, Message], str]]):
    """
    The default message function factory.

    This provides a callable that will fire a command if the message begins with the specified
    prefix or list of prefixes.

    If ``command_prefix`` is provided to the :class:`.Client`, then it will automatically call this
    function to get a message check function to use.

    .. code-block:: python3

        # verbose form
        message_check = prefix_check_factory(["!", "?"])
        cl = Client(message_check=message_check)

        # implicit form
        cl = Client(command_prefix=["!", "?"])

    The :attr:`prefix` is set on the returned function that can be used to retrieve the prefixes
    defined to create  the function at any time.

    :param prefix: A :class:`str` or :class:`typing.Iterable[str]` that represents the prefix(es) \
        to use.
    :return: A callable that can be used for the ``message_check`` function on the client.
    """

    async def __inner(bot: Client, message: Message):
        # move prefix out of global scope
        _prefix = prefix
        matched = None

        if callable(_prefix):
            _prefix = _prefix(bot, message)
            if inspect.isawaitable(_prefix):
                _prefix = await _prefix

        if isinstance(_prefix, str):
            match = message.content.startswith(_prefix)
            if match:
                matched = _prefix

        elif isinstance(prefix, collections.Iterable):
            for i in _prefix:
                if message.content.startswith(i):
                    matched = i
                    break

        if not matched:
            return None

        tokens = split_message_content(message.content[len(matched):])
        command_word = tokens[0]

        return command_word, tokens[1:]

    __inner.prefix = prefix
    return __inner
